import sys
from enum import Enum
from typing import TYPE_CHECKING

from mstrio.access_and_security.security_role import SecurityRole  # noqa: F401
from mstrio.datasources import DatasourceConnection  # noqa: F401
from mstrio.datasources import DatasourceInstance, DatasourceLogin  # noqa: F401
from mstrio.distribution_services.schedule import Schedule  # noqa: F401
from mstrio.modeling.filter import Filter  # noqa: F401
from mstrio.modeling.metric import Metric  # noqa: F401
from mstrio.modeling.schema import (  # noqa: F401
    Attribute,
    Fact,
    LogicalTable,
    Transformation,
    UserHierarchy,
)
from mstrio.modeling.security_filter import SecurityFilter  # noqa: F401
from mstrio.object_management.folder import Folder  # noqa: F401
from mstrio.object_management.object import Object
from mstrio.object_management.search_operations import SearchObject  # noqa: F401
from mstrio.object_management.shortcut import Shortcut  # noqa: F401
from mstrio.project_objects import Document, Report  # noqa: F401
from mstrio.project_objects.datasets import OlapCube, SuperCube  # noqa: F401
from mstrio.server import Project  # noqa: F401
from mstrio.types import ObjectSubTypes, ObjectTypes
from mstrio.users_and_groups import User, UserGroup  # noqa: F401

if TYPE_CHECKING:
    from mstrio.connection import Connection


class TypeObjectMapping(Enum):
    Report = ObjectTypes.REPORT_DEFINITION  # noqa: F811
    Folder = ObjectTypes.FOLDER  # noqa: F811
    DatasourceInstance = ObjectTypes.DBROLE  # noqa: F811
    DatasourceLogin = ObjectTypes.DBLOGIN  # noqa: F811
    DatasourceConnection = ObjectTypes.DBCONNECTION  # noqa: F811
    Project = ObjectTypes.PROJECT  # noqa: F811
    SecurityRole = ObjectTypes.SECURITY_ROLE  # noqa: F811
    Schedule = ObjectTypes.SCHEDULE_TRIGGER  # noqa: F811
    Document = ObjectTypes.DOCUMENT_DEFINITION  # noqa: F811
    SecurityFilter = ObjectTypes.SECURITY_FILTER  # noqa: F811
    SearchObject = ObjectTypes.SEARCH  # noqa: F811
    Attribute = ObjectTypes.ATTRIBUTE  # noqa: F811
    Filter = ObjectTypes.FILTER  # noqa: F811
    LogicalTable = ObjectTypes.TABLE  # noqa: F811
    Fact = ObjectTypes.FACT  # noqa: F811
    Transformation = ObjectTypes.ROLE  # noqa: F811
    Metric = ObjectTypes.METRIC  # noqa: F811
    UserHierarchy = ObjectTypes.DIMENSION  # noqa: F811
    Shortcut = ObjectTypes.SHORTCUT_TYPE  # noqa: F811


class SubTypeObjectMapping(Enum):
    Filter = ObjectSubTypes.FILTER  # noqa: F811
    OlapCube = ObjectSubTypes.OLAP_CUBE  # noqa: F811
    SuperCube = ObjectSubTypes.SUPER_CUBE  # noqa: F811
    User = ObjectSubTypes.USER  # noqa: F811
    UserGroup = ObjectSubTypes.USER_GROUP  # noqa: F811


def __str_to_class(classname):
    return getattr(sys.modules[__name__], classname)


def map_to_object(
    object_type: int | ObjectTypes, subtype: int | ObjectSubTypes | None = None
):
    if not isinstance(object_type, ObjectTypes):
        try:
            object_type = ObjectTypes(object_type)
        except ValueError:
            object_type = ObjectTypes.NOT_SUPPORTED
    if not isinstance(subtype, ObjectSubTypes):
        try:
            subtype = ObjectSubTypes(subtype)
        except ValueError:
            subtype = ObjectSubTypes.NOT_SUPPORTED

    if (
        object_type == ObjectTypes.REPORT_DEFINITION
        and subtype in (ObjectSubTypes.OLAP_CUBE, ObjectSubTypes.SUPER_CUBE)
    ) or (
        object_type == ObjectTypes.USER
        and subtype in (ObjectSubTypes.USER, ObjectSubTypes.USER_GROUP)
    ):
        return __str_to_class(SubTypeObjectMapping(subtype).name)
    elif object_type in [v.value for v in TypeObjectMapping]:
        return __str_to_class(TypeObjectMapping(object_type).name)
    else:
        return Object


def map_object(connection: "Connection", obj: dict):
    """Map a dict that represents an object to an instance of the corresponding
    mstrio class.
    """

    return map_to_object(obj.get('type'), obj.get('subtype')).from_dict(
        source=obj, connection=connection
    )


def map_objects_list(connection: "Connection", objects_list: list):
    """Map a list of dict that represent objects to instances of
    the corresponding mstrio classes.
    """

    return [map_object(connection, obj) for obj in objects_list]



import csv
import json
import logging
import pickle
from abc import ABCMeta, abstractclassmethod
from ast import literal_eval
from typing import TYPE_CHECKING

from mstrio import config
from mstrio.helpers import VersionException
from mstrio.utils import helper

from .settings_helper import convert_settings_to_mega_byte

if TYPE_CHECKING:
    from mstrio.utils.settings.base_settings import BaseSettings

logger = logging.getLogger(__name__)

EXPORTED_MSG = "Settings exported to"


class SettingsSerializerFactory:
    def get_serializer(self, file):
        file_type = None
        for extension in ['.json', '.csv', '.p', '.pkl', '.pickle']:
            if extension in file[-7:].lower():
                file_type = extension[1:]

        if file_type is None:
            raise ValueError(
                "This file type is not supported. Supported file types are .json, .csv,"
                " .pickle"
            )
        elif file_type == 'json':
            return JSONSettingsIO()
        elif file_type == 'csv':
            return CSVSettingsIO()
        elif file_type in ['pkl', 'pickle', 'p']:
            return PickleSettingsIO()


class SettingsIO(metaclass=ABCMeta):
    FILE_NAME: str | tuple | None = None
    PROJECT_VERSION: int | None = None
    SERVER_VERSION: int | None = None

    @abstractclassmethod
    def to_file(cls, file: str, settings_obj: "BaseSettings") -> None:
        pass

    @abstractclassmethod
    def from_file(cls, file: str, settings_obj: "BaseSettings") -> dict:
        pass

    @classmethod
    def validate_file_name(cls, file):
        if not isinstance(cls.FILE_NAME, (str, tuple)):
            raise ValueError("FILE_NAME not defined")
        elif not file.endswith(cls.FILE_NAME):
            msg = (
                f"The file extension is different than {cls.FILE_NAME}, please note "
                f"that using a different extension might disrupt opening the file "
                f"correctly."
            )
            helper.exception_handler(msg, exception_type=Warning)

    @classmethod
    def check_type(
        cls, settings_type: str | None, settings_obj: "BaseSettings"
    ) -> None:
        if settings_type is None:
            return None
        elif settings_type != settings_obj._TYPE:
            raise TypeError('Unsupported settings.')

    @classmethod
    def check_version(
        cls, version: str | int | None, settings_type: str | None
    ) -> None:
        if isinstance(version, str):
            version = int(version)
        if version is None or settings_type is None:
            return None
        else:
            server_settings_not_supported = (
                settings_type == "allServerSettings" and version > cls.SERVER_VERSION
            )
            project_settings_not_supported = (
                settings_type == "allProjectSettings" and version > cls.PROJECT_VERSION
            )
            if server_settings_not_supported or project_settings_not_supported:
                raise VersionException("Unsupported settings version")

    @classmethod
    def get_version(cls, settings_obj: "BaseSettings") -> int:
        if settings_obj._TYPE == "allServerSettings":
            return cls.SERVER_VERSION
        elif settings_obj._TYPE == "allProjectSettings":
            return cls.PROJECT_VERSION


class PickleSettingsIO(SettingsIO):
    FILE_NAME = ('.pkl', '.pickle', '.p')
    PROJECT_VERSION = 2
    SERVER_VERSION = 2

    @classmethod
    def to_file(cls, file: str, settings_obj: "BaseSettings") -> None:
        cls.validate_file_name(file)
        settings_dict = settings_obj.list_properties(show_names=False)

        with open(file, 'wb') as f:
            # save version and type
            settings_dict.update(
                __version__=cls.get_version(settings_obj), __page__=settings_obj._TYPE
            )
            pickle.dump(
                settings_dict, f, protocol=4
            )  # pickle protocol added in python 3.4
        if config.verbose:
            logger.info(f"{EXPORTED_MSG} '{file}'")

    @classmethod
    def from_file(cls, file: str, settings_obj: "BaseSettings") -> dict:
        with open(file, 'rb') as f:
            settings_dict = pickle.load(f)

            # check versioning and type
            version = settings_dict.pop("__version__", None)
            settings_type = settings_dict.pop('__page__', None)
            cls.check_type(settings_type, settings_obj)
            cls.check_version(version, settings_type)
            # convert memory units for v1 (no version) settings
            if version is None:
                settings_dict = convert_settings_to_mega_byte(
                    settings_dict, settings_obj._CONVERSION_MAP
                )

        return settings_dict


class JSONSettingsIO(SettingsIO):
    FILE_NAME = '.json'
    PROJECT_VERSION = 2
    SERVER_VERSION = 2

    @classmethod
    def to_file(cls, file: str, settings_obj: "BaseSettings") -> None:
        cls.validate_file_name(file)
        settings_dict = settings_obj.list_properties(show_names=False)

        with open(file, 'w') as f:
            # save version and type
            settings_dict.update(
                __version__=cls.get_version(settings_obj), __page__=settings_obj._TYPE
            )
            json.dump(settings_dict, f, indent=4)
        if config.verbose:
            logger.info(f"{EXPORTED_MSG} '{file}'")

    @classmethod
    def from_file(cls, file: str, settings_obj: "BaseSettings") -> dict:
        with open(file) as f:
            settings_dict = json.load(f)

            # check versioning and type
            version = settings_dict.pop("__version__", None)
            settings_type = settings_dict.pop('__page__', None)
            cls.check_type(settings_type, settings_obj)
            cls.check_version(version, settings_type)
            # convert memory units for v1 (no version) settings
            if version is None:
                settings_dict = convert_settings_to_mega_byte(
                    settings_dict, settings_obj._CONVERSION_MAP
                )

        return settings_dict


class CSVSettingsIO(SettingsIO):
    FILE_NAME = '.csv'
    PROJECT_VERSION = 1
    SERVER_VERSION = 2

    @classmethod
    def to_file(
        cls, file: str, settings_obj: "BaseSettings", show_description: bool = False
    ) -> None:
        cls.validate_file_name(file)
        # newline = '' to disable universal newlines translation
        # It should always be safe to specify newline='', since the csv module
        # does its own (universal) newline handling.
        # https://docs.python.org/3/library/csv.html#examples
        with open(file, 'w', newline='') as f:
            # Add lines for workstation compatibility
            version = cls.get_version(settings_obj)
            fieldnames = [
                'Name',
                'Value',
                *(('Description',) if show_description else ()),
            ]
            f.write(
                f"#__page__,{settings_obj._TYPE}\n"
                f"#__version__,{version}\n"
                f"{','.join(fieldnames)}\n"
            )
            w = csv.DictWriter(f, fieldnames=fieldnames, quoting=csv.QUOTE_ALL)

            settings_dict = settings_obj.list_properties(
                show_names=False, show_description=show_description
            )
            rows = [
                {
                    'Name': setting,
                    'Value': value.get('value') if show_description else value,
                    **(
                        {'Description': value.get('description')}
                        if show_description
                        else {}
                    ),
                }
                for setting, value in settings_dict.items()
            ]
            w.writerows(rows)
            if config.verbose:
                logger.info(f"{EXPORTED_MSG} '{file}'")

    @classmethod
    def from_file(cls, file: str, settings_obj: "BaseSettings") -> dict:
        with open(file) as f:
            settings_dict = {
                # Add only name and value and ignore description
                name: value
                for name, value, *desc in csv.reader(f, quoting=csv.QUOTE_ALL)
            }
            return cls.process_csv_settings(settings_dict, settings_obj)

    @classmethod
    def process_csv_settings(cls, settings: dict, settings_obj: "BaseSettings") -> dict:
        """Helper function to extract settings from csv settings files."""
        version = settings.pop("#__version__", None)
        settings_type = settings.pop('#__page__', None)
        cls.check_type(settings_type, settings_obj)
        cls.check_version(version, settings_type)

        processed_settings = {}
        for setting, value in settings.items():
            if setting == 'Name':  # skip header line
                continue
            value_lower = value.lower()
            if value_lower == 'false':
                processed_value = False
            elif value_lower == 'true':
                processed_value = True
            elif value.startswith('['):
                processed_value = literal_eval(value)
            elif not bool(value and value.strip()):
                processed_value = value
            else:
                processed_value = settings_obj._cast_type_from_obj(setting, value)

            processed_settings.update({setting: processed_value})

        # convert memory units for v1 Server settings
        if version == "1" and settings_type == "allServerSettings":
            processed_settings = convert_settings_to_mega_byte(
                processed_settings, settings_obj._CONVERSION_MAP
            )

        return processed_settings

    @classmethod
    def check_type(
        cls, settings_type: str | None, settings_obj: "BaseSettings"
    ) -> None:
        if settings_type is None:
            raise ValueError('CSV settings are missing `#__page__` header')
        elif settings_type != settings_obj._TYPE:
            raise TypeError('Unsupported settings.')

    @classmethod
    def check_version(cls, version: int | None, settings_type: str) -> None:
        if version is None:
            raise ValueError('CSV settings are missing `#__version__` header')
        elif isinstance(version, str):
            version = int(version)

        server_settings_not_supported = (
            settings_type == "allServerSettings" and version > cls.SERVER_VERSION
        )
        project_settings_not_supported = (
            settings_type == "allProjectSettings" and version > cls.PROJECT_VERSION
        )
        if server_settings_not_supported or project_settings_not_supported:
            raise VersionException('Unsupported CSV settings version.')



from typing import TYPE_CHECKING

from requests import Response

from mstrio.utils.error_handlers import ErrorHandler
from mstrio.utils.helper import deprecation_warning, get_valid_project_id

if TYPE_CHECKING:
    from mstrio.connection import Connection
    from mstrio.utils.sessions import FuturesSessionWithRenewal


@ErrorHandler(err_msg="Error getting available dossiers.")
def get_dossiers(
    connection: 'Connection',
    offset: int = 0,
    limit: int = -1,
    search_term: str | None = None,
    certified_status: str | None = None,
    search_pattern: str | None = None,
    fields: str | None = None,
    project_id: str | None = None,
    error_msg: str | None = None,
) -> Response:
    """Get the list of available dossiers.

    Args:
        connection(object): MicroStrategy REST API connection object
        offset(int): Starting point within the collection of returned search
            results. Used to control paging behavior.
        limit(int): Maximum number of items returned for a single request.
            Used to control paging behavior.
        search_term(str, optional): The value that the search searchPattern is
            set to.
        certified_status(str, optional): Define a search criteria of the
            certified status of the object
        search_pattern(str, optional): The kind of search pattern that will be
            applied to the search,
        fields(str, optional): A whitelist of top-level fields separated by
            commas. Allow the client to selectively retrieve fields in the
            response.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    endpoint = '/api/dossiers'
    return connection.get(
        endpoint=endpoint,
        headers={'X-MSTR-ProjectID': project_id},
        params={
            'searchTerm': search_term,
            'searchPattern': search_pattern,
            'offset': offset,
            'limit': limit,
            'certifiedStatus': certified_status,
            'fields': fields,
        },
    )


@ErrorHandler(err_msg="Error getting available dashboards.")
def get_dashboards(
    connection: 'Connection',
    offset: int = 0,
    limit: int = -1,
    search_term: str | None = None,
    certified_status: str | None = None,
    search_pattern: str | None = None,
    fields: str | None = None,
    project_id: str | None = None,
    error_msg: str | None = None,
) -> Response:
    """Get the list of available dashboards.

    Args:
        connection(object): MicroStrategy REST API connection object
        offset(int): Starting point within the collection of returned search
            results. Used to control paging behavior.
        limit(int): Maximum number of items returned for a single request.
            Used to control paging behavior.
        search_term(str, optional): The value that the search searchPattern is
            set to.
        certified_status(str, optional): Define a search criteria of the
            certified status of the object
        search_pattern(str, optional): The kind of search pattern that will be
            applied to the search,
        fields(str, optional): A whitelist of top-level fields separated by
            commas. Allow the client to selectively retrieve fields in the
            response.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    endpoint = '/api/dossiers'
    return connection.get(
        endpoint=endpoint,
        headers={'X-MSTR-ProjectID': project_id},
        params={
            'searchTerm': search_term,
            'searchPattern': search_pattern,
            'offset': offset,
            'limit': limit,
            'certifiedStatus': certified_status,
            'fields': fields,
        },
    )


def get_dossiers_async(
    future_session: 'FuturesSessionWithRenewal',
    offset: int = 0,
    limit: int = -1,
    search_term: str | None = None,
    certified_status: str | None = None,
    search_pattern: str | None = None,
    fields: str | None = None,
    project_id: str | None = None,
):
    """Get the list of available dossiers asynchronously.

    Args:
        future_session: FuturesSessionWithRenewal object
        offset(int): Starting point within the collection of returned search
            results. Used to control paging behavior.
        limit(int): Maximum number of items returned for a single request.
            Used to control paging behavior.
        search_term(str, optional): The value that the search searchPattern is
            set to.
        certified_status(str, optional): Define a search criteria of the
            certified status of the object
        search_pattern(str, optional): The kind of search pattern that will be
            applied to the search,
        fields(str, optional): A whitelist of top-level fields separated by
            commas. Allow the client to selectively retrieve fields in the
            response.
        project_id(str, optional): Project ID of the project to use. If not
            set then the project selected in `connection` will be used.

    Returns:
        Complete HTTP response object.
    """
    future_session.connection._validate_project_selected()
    endpoint = '/api/dossiers'
    headers = ({'X-MSTR-ProjectID': project_id},)
    params = {
        'searchTerm': search_term,
        'searchPattern': search_pattern,
        'offset': offset,
        'limit': limit,
        'certifiedStatus': certified_status,
        'fields': fields,
    }
    return future_session.get(endpoint=endpoint, params=params, headers=headers)


def get_dashboards_async(
    future_session: 'FuturesSessionWithRenewal',
    offset: int = 0,
    limit: int = -1,
    search_term: str | None = None,
    certified_status: str | None = None,
    search_pattern: str | None = None,
    fields: str | None = None,
    project_id: str | None = None,
):
    """Get the list of available dashboards asynchronously.

    Args:
        future_session: FuturesSessionWithRenewal object
        offset(int): Starting point within the collection of returned search
            results. Used to control paging behavior.
        limit(int): Maximum number of items returned for a single request.
            Used to control paging behavior.
        search_term(str, optional): The value that the search searchPattern is
            set to.
        certified_status(str, optional): Define a search criteria of the
            certified status of the object
        search_pattern(str, optional): The kind of search pattern that will be
            applied to the search,
        fields(str, optional): A whitelist of top-level fields separated by
            commas. Allow the client to selectively retrieve fields in the
            response.
        project_id(str, optional): Project ID of the project to use. If not
            set then the project selected in `connection` will be used.

    Returns:
        Complete HTTP response object.
    """
    future_session.connection._validate_project_selected()
    endpoint = '/api/dossiers'
    headers = ({'X-MSTR-ProjectID': project_id},)
    params = {
        'searchTerm': search_term,
        'searchPattern': search_pattern,
        'offset': offset,
        'limit': limit,
        'certifiedStatus': certified_status,
        'fields': fields,
    }
    return future_session.get(endpoint=endpoint, params=params, headers=headers)


@ErrorHandler(err_msg="Error getting available documents.")
def get_documents(
    connection: 'Connection',
    offset: int = 0,
    limit: int = -1,
    search_pattern: str = None,
    search_term: str | None = None,
    certified_status: str | None = None,
    project_id: str | None = None,
    fields: str | None = None,
    error_msg: str | None = None,
) -> Response:
    """Get the list of available documents.

    Args:
        connection: MicroStrategy REST API connection object
        offset(int): Starting point within the collection of returned search
            results. Used to control paging behavior.
        limit(int): Maximum number of items returned for a single request.
            Used to control paging behavior.
        search_pattern(str, optional): The kind of search pattern that will be
            applied to the search,
        search_term(str, optional): The value that the search searchPattern is
            set to.
        certified_status(str, optional): Define a search criteria of the
            certified status of the object
        project_id(str, optional): Project ID
        fields(str, optional): A whitelist of top-level fields separated by
            commas. Allow the client to selectively retrieve fields in the
            response.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    endpoint = '/api/documents/'
    return connection.get(
        endpoint=endpoint,
        headers={'X-MSTR-ProjectID': project_id},
        params={
            'searchTerm': search_term,
            'searchPattern': search_pattern,
            'offset': offset,
            'limit': limit,
            'certifiedStatus': certified_status,
            'fields': fields,
        },
    )


def get_documents_async(
    future_session: 'FuturesSessionWithRenewal',
    offset: int = 0,
    limit: int = -1,
    search_pattern: str = None,
    search_term: str | None = None,
    project_id: str | None = None,
    certified_status: str | None = None,
    fields: str | None = None,
):
    """Get the list of available documents asynchronously.

    Args:
        future_session: FutureSession object
        offset(int): Starting point within the collection of returned search
            results. Used to control paging behavior.
        limit(int): Maximum number of items returned for a single request.
            Used to control paging behavior.
        search_pattern(str, optional): The kind of search pattern that will be
            applied to the search,
        search_term(str, optional): The value that the search searchPattern is
            set to.
        project_id(str, optional): Project ID
        certified_status(str, optional): Define a search criteria of the
            certified status of the object
        fields(str, optional): A whitelist of top-level fields separated by
            commas. Allow the client to selectively retrieve fields in the
            response.

    Returns:
        Complete HTTP response object.
    """
    endpoint = '/api/documents/'
    params = {
        'searchTerm': search_term,
        'searchPattern': search_pattern,
        'offset': offset,
        'limit': limit,
        'certifiedStatus': certified_status,
        'fields': fields,
    }
    headers = {'X-MSTR-ProjectID': project_id}
    future = future_session.get(endpoint=endpoint, params=params, headers=headers)
    return future


@ErrorHandler(err_msg="Error getting document {document_id}")
def get_document_status(connection, document_id, instance_id, error_msg=None):
    """Get the status of a dashboard, document or dossier instance.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        instance_id (string): Document Instance ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint=f'api/documents/{document_id}/instances/{instance_id}/status',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg="Error getting prompts for document {document_id}")
def get_prompts_for_instance(connection, document_id, instance_id, error_msg=None):
    """Get the collection of prompts and their respective definitions from a
    dashboard/document/dossier instance.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        instance_id (string): Document Instance ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    endpoint = f'/api/documents/{document_id}/instances/{instance_id}/prompts'
    return connection.get(endpoint=endpoint, headers={'X-MSTR-ProjectID': None})


@ErrorHandler(err_msg="Error getting attribute element for prompt {prompt_identifier}")
def get_attribute_element_for_prompt(
    connection, document_id, instance_id, prompt_identifier, error_msg=None
):
    """Get available attribute element for dashboard/document/dossier's
    attribute element prompt.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        instance_id (string): Document Instance ID
        prompt_identifier (string): Prompt key or ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    endpoint = (
        f'/api/documents/{document_id}/instances/{instance_id}'
        f'/prompts/{prompt_identifier}/elements'
    )
    return connection.get(endpoint=endpoint, headers={'X-MSTR-ProjectID': None})


@ErrorHandler(err_msg="Error getting available object for prompt {prompt_identifier}")
def get_available_object(
    connection, document_id, instance_id, prompt_identifier, error_msg=None
):
    """Get available object for answering all kinds of prompts.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        instance_id (string): Document Instance ID
        prompt_identifier (string): Prompt key or ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    endpoint = (
        f'api/documents/{document_id}/instances/'
        f'{instance_id}/prompts/{prompt_identifier}/objects'
    )
    return connection.get(endpoint=endpoint, headers={'X-MSTR-ProjectID': None})


@ErrorHandler(
    err_msg="Error exporting visualization for document {document_id} to PDF file."
)
def export_visualization_to_pdf(
    connection, document_id, instance_id, node_key, body, error_msg=None
):
    """Export a single visualization from a specific document instance to a PDF
    file.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        instance_id (string): Document Instance ID
        body: JSON-formatted information used to format the document
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    endpoint = (
        f'/api/documents/{document_id}/instances/{instance_id}'
        f'/visualizations/{node_key}/pdf'
    )
    return connection.post(
        endpoint=endpoint, headers={'X-MSTR-ProjectID': None}, json=body
    )


@ErrorHandler(
    err_msg="Error exporting visualization for document {document_id} to CSV file."
)
def export_visualization_to_csv(
    connection, document_id, instance_id, node_key, body, error_msg=None
):
    """Export a single visualization from a specific document instance to a CSV
    file.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        instance_id (string): Document Instance ID
        node_key (string): Visualization node key
        body: JSON-formatted information used to format the document
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    endpoint = (
        f'/api/documents/{document_id}/instances/{instance_id}'
        f'/visualizations/{node_key}/csv'
    )
    return connection.post(
        endpoint=endpoint, headers={'X-MSTR-ProjectID': None}, json=body
    )


@ErrorHandler(err_msg="Error exporting document {document_id} to PDF file.")
def export_document_to_pdf(
    connection: 'Connection',
    document_id: str,
    instance_id: str,
    body: dict,
    error_msg: str | None = None,
) -> Response:
    """Export a specific document instance to a PDF file.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        instance_id (string): Document Instance ID
        body: JSON-formatted information used to format the document
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    project_id = get_valid_project_id(
        connection=connection, project_id=connection.project_id
    )
    endpoint = f'/api/documents/{document_id}/instances/{instance_id}/pdf'
    return connection.post(
        endpoint=endpoint, headers={'X-MSTR-ProjectID': project_id}, json=body
    )


@ErrorHandler(err_msg="Error exporting document {document_id} to .mstr file")
def export_document_to_mstr(
    connection: 'Connection',
    document_id: str,
    instance_id: str,
    body: dict,
    error_msg: str | None = None,
) -> Response:
    """Export a specific document in a specific project to an .mstr file.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        instance_id (string): Document Instance ID
        body: JSON-formatted information used to format the document
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    project_id = get_valid_project_id(
        connection=connection, project_id=connection.project_id
    )
    endpoint = f'/api/documents/{document_id}/instances/{instance_id}/mstr'
    return connection.post(
        endpoint=endpoint, headers={'X-MSTR-ProjectID': project_id}, json=body
    )


@ErrorHandler(err_msg="Error exporting document {document_id} to Excel file.")
def export_document_to_excel(
    connection: 'Connection',
    document_id: str,
    instance_id: str,
    body: dict,
    error_msg: str | None = None,
) -> Response:
    """Export a document from a specific document instance to an Excel file.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        instance_id (string): Document Instance ID
        body: JSON-formatted information used to format the document
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    project_id = get_valid_project_id(
        connection=connection, project_id=connection.project_id
    )
    endpoint = f'/api/documents/{document_id}/instances/{instance_id}/excel'
    return connection.post(
        endpoint=endpoint, headers={'X-MSTR-ProjectID': project_id}, json=body
    )


@ErrorHandler(err_msg="Error setting document {document_id} to prompt status.")
def set_document_to_prompt_status(connection, document_id, instance_id, error_msg=None):
    """Export a document from a specific document instance to an Excel file.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        instance_id (string): Document Instance ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    endpoint = f'/api/documents/{document_id}/instances/{instance_id}/rePrompt'
    return connection.post(endpoint=endpoint, headers={'X-MSTR-ProjectID': None})


@ErrorHandler(err_msg="Error getting cubes used by document {document_id}")
def get_cubes_used_by_document(connection, document_id, error_msg=None):
    """Get the cubes used by a document in a specific project, either directly
    or indirectly.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    endpoint = f'/api/documents/{document_id}/cubes'
    return connection.get(endpoint=endpoint)


@ErrorHandler(err_msg="Error overwriting document {document_id}")
def overwrite_document(connection, document_id, instance_id, error_msg=None):
    """Save a document instance by overwriting an existing document.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        instance_id (string): Document Instance ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    endpoint = f'/api/documents/{document_id}/instances/{instance_id}/save'
    return connection.post(endpoint=endpoint, headers={'X-MSTR-ProjectID': None})


@ErrorHandler(err_msg="Error saving document {document_id}")
def save_document_as(connection, document_id, instance_id, error_msg=None):
    """Save a document instance by creating a new document.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        instance_id (string): Document Instance ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    endpoint = f'/api/documents/{document_id}/instances/{instance_id}/saveAs'
    return connection.post(endpoint=endpoint, headers={'X-MSTR-ProjectID': None})


@ErrorHandler(err_msg="Error creating instance for document {document_id}")
def create_new_document_instance(connection, document_id, body, error_msg=None):
    """Execute a specific document in a specific project and create an instance
    of the document.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    endpoint = f'/api/documents/{document_id}/instances'
    return connection.post(endpoint=endpoint, json=body)


@ErrorHandler(err_msg="Error deleting document instance {instance_id}")
def delete_document_instance(connection, document_id, instance_id, error_msg=None):
    """Save a document instance by overwriting an existing document.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        instance_id (string): Document Instance ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    endpoint = f'/api/documents/{document_id}/instances/{instance_id}'
    return connection.delete(endpoint=endpoint, headers={'X-MSTR-ProjectID': None})


@ErrorHandler(err_msg="Error refreshing document instance {instance_id}")
def refresh_document_instance(connection, document_id, instance_id, error_msg=None):
    """Execute a specific document in a specific project and create an instance
    of the document.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        instance_id (string): Document Instance ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    endpoint = f'/api/documents/{document_id}/instances/{instance_id}/refresh'
    return connection.put(endpoint=endpoint, headers={'X-MSTR-ProjectID': None})


@ErrorHandler(err_msg="Error getting collection of prompts for document {document_id}")
def get_prompts(connection, document_id, error_msg=None):
    """Get the collection of prompts and their respective definitions from a
    dashboard/document/dossier definition.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    endpoint = f'/api/documents/{document_id}/prompts'
    return connection.get(endpoint=endpoint, headers={'X-MSTR-ProjectID': None})


@ErrorHandler(err_msg="Error answering prompt for instance {instance_id}")
def answer_prompts(connection, document_id, instance_id, body, error_msg=None):
    """Answer specified prompts on the dashboard/document/dossier instance,
    prompts can either be answered with default answers(if available),
    the appropriate answers, or if the prompt is not required
    the prompt can simply be closed.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        instance_id (string): Document Instance ID
        body: JSON-formatted information used to format the document
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    endpoint = f'/api/documents/{document_id}/instances/{instance_id}/prompts/answers'
    return connection.put(
        endpoint=endpoint, headers={'X-MSTR-ProjectID': None}, json=body
    )


@ErrorHandler(err_msg="Error retrieving document shortcut for document {document_id}")
def get_document_shortcut(connection, document_id, instance_id, error_msg=None):
    """Retrieve a published shortcut from a specific document instance.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document ID
        instance_id (string): Document Instance ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    endpoint = f'/api/documents/{document_id}/instances/{instance_id}/shortcut'
    return connection.get(endpoint=endpoint)


@ErrorHandler(err_msg="Error creating instance for dossier {dossier_id}")
def create_dossier_instance(connection, dossier_id, body, error_msg=None):
    """Execute a specific dossier and create an instance of the dossier.

    Args:
        connection: MicroStrategy REST API connection object
        dossier_id (string): Dossier ID
        body: JSON-formatted information used to format the document
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    endpoint = f'/api/dossiers/{dossier_id}/instances'
    return connection.post(
        endpoint=endpoint, headers={'X-MSTR-ProjectID': None}, json=body
    )


@ErrorHandler(err_msg="Error creating instance for dashboard {dashboard_id}")
def create_dashboard_instance(connection, dashboard_id, body, error_msg=None):
    """Execute a specific dossier and create an instance of the dossier.

    Args:
        connection: MicroStrategy REST API connection object
        dashboard_id (string): Dashboard ID
        body: JSON-formatted information used to format the document
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    endpoint = f'/api/dossiers/{dashboard_id}/instances'
    return connection.post(
        endpoint=endpoint, headers={'X-MSTR-ProjectID': None}, json=body
    )


@ErrorHandler(err_msg="Error getting hierarchy for dossier {id}")
def get_dossier_hierarchy(connection: 'Connection', id: str) -> Response:
    """Get the hierarchy of a specific dossier in a specific project.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): Dossier ID

    Returns:
        Complete HTTP response object.
    """
    endpoint = f'/api/v2/dossiers/{id}/definition/'
    return connection.get(endpoint=endpoint)


@ErrorHandler(err_msg="Error getting hierarchy for dashboard {id}")
def get_dashboard_hierarchy(connection: 'Connection', id: str) -> Response:
    """Get the hierarchy of a specific dossier in a specific project.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): Dossier ID

    Returns:
        Complete HTTP response object.
    """
    endpoint = f'/api/v2/dossiers/{id}/definition/'
    return connection.get(endpoint=endpoint)


@ErrorHandler(err_msg="Error getting definition of document {id}")
def get_document_definition(connection, id, error_msg=None):
    """Get the hierarchy of a specific dossier in a specific project.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): Dossier ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    endpoint = f'/api/v2/documents/{id}'
    return connection.get(endpoint=endpoint, headers={'X-MSTR-ProjectID': None})


@ErrorHandler(err_msg="Error getting dossier hierarchy from instance {instance_id}")
def get_dossier_hierarchy_from_instance(
    connection, dossier_id, instance_id, error_msg=None
):
    """Get the hierarchy of a specific dossier in a specific project from
    instance.

    Args:
        connection: MicroStrategy REST API connection object
        dossier_id (string): Dossier ID
        instance_id (string): Document Instance ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    endpoint = f'api/v2/dossiers/{dossier_id}/instances/{instance_id}/definition'
    return connection.get(endpoint=endpoint, headers={'X-MSTR-ProjectID': None})


@ErrorHandler(err_msg="Error getting dossier hierarchy from instance {instance_id}")
def get_dashboard_hierarchy_from_instance(
    connection, dashboard_id, instance_id, error_msg=None
):
    """Get the hierarchy of a specific dashboard in a specific project from
    instance.

    Args:
        connection: MicroStrategy REST API connection object
        dashboard_id (string): Dashboard ID
        instance_id (string): Document Instance ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    endpoint = f'api/v2/dossiers/{dashboard_id}/instances/{instance_id}/definition'
    return connection.get(endpoint=endpoint, headers={'X-MSTR-ProjectID': None})


@ErrorHandler(
    err_msg="Error getting definition and results for dashboard/dossier {dossier_id}"
)
def get_definition_and_results_of_visualization(
    connection,
    dossier_id,
    instance_id,
    chapter_key,
    visualization_key,
    dashboard_id=None,
    error_msg=None,
):
    """Get the definition and data result of a grid/graph visualization in a
    specific dashboard/dossier in a specific project.

    Args:
        connection: MicroStrategy REST API connection object
        dossier_id (string): Dossier ID
        instance_id (string): Document Instance ID
        chapter_key (string): Chapter Key
        visualization_key (string): Visualization Key
        dashboard_id (string, optional): Dashboard ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    if dossier_id:
        dashboard_id = dossier_id
        deprecation_warning(
            "argument `dossier_id`", "argument `dashboard_id`", "11.5.03", False
        )

    endpoint = (
        f'/api/v2/dossiers/{dashboard_id}'
        f'/instances/{instance_id}/chapters/{chapter_key}'
        f'/visualizations/{visualization_key}'
    )
    return connection.get(endpoint=endpoint, headers={'X-MSTR-ProjectID': None})



from typing import TYPE_CHECKING, Union

from pandas import DataFrame

from mstrio.api import security
from mstrio.connection import Connection
from mstrio.users_and_groups.user_group import UserGroup
from mstrio.utils import helper
from mstrio.utils.entity import EntityBase
from mstrio.utils.version_helper import class_version_handler

if TYPE_CHECKING:
    from mstrio.users_and_groups.user import User


@class_version_handler('11.2.0100')
class Privilege(EntityBase):
    """Object representation of MicroStrategy Privilege object.

    Attributes:
        connection: A MicroStrategy connection object
        id: privilege ID
        name: privilege name
        description: privilege description
        categories: privilege category
        is_project_level_privilege: specify if privilege is compatible with
            server configuration level or project level
    """

    def __init__(
        self, connection: Connection, name: str | None = None, id: str | None = None
    ) -> None:
        """Initialize Privilege object by passing `name` or `id`. When `id` is
        provided (not `None`), `name` is omitted. To explore all available
        privileges use the `list_privileges()` method.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
            name: exact name of privilege
            id: ID of privilege
        """
        if name is None and id is None:
            helper.exception_handler(
                "Please specify either 'name' or 'id' parameter in the constructor.",
                exception_type=ValueError,
            )

        if name is None or (name and id):
            privileges = Privilege.list_privileges(
                connection=connection, to_dictionary=True, id=str(id)
            )
            if privileges:
                [privilege] = privileges
                self._set_object_attributes(**privilege)
            else:
                helper.exception_handler(
                    f"There is no Privilege with the given id: '{id}'",
                    exception_type=ValueError,
                )
        if id is None:
            privileges = Privilege.list_privileges(
                connection=connection, to_dictionary=True, name=name
            )
            if privileges:
                [privilege] = privileges
                self._set_object_attributes(**privilege)
            else:
                helper.exception_handler(
                    f"There is no Privilege with the given name: '{name}'",
                    exception_type=ValueError,
                )
        super().__init__(connection, self.id, name=self.name)

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self._is_project_level_privilege = kwargs.get("is_project_level_privilege")
        self._categories = kwargs.get("categories")

    @classmethod
    def list_privileges(
        cls,
        connection: Connection,
        to_dictionary: bool = False,
        to_dataframe: bool = False,
        **filters,
    ) -> list['Privilege'] | list[dict] | DataFrame:
        """Get list of privilege objects or privilege dicts. Filter the
        privileges by specifying the `filters` keyword arguments.

        Optionally use `to_dictionary` or `to_dataframe` to choose output
        format.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
            to_dictionary: If `True` returns dict, by default (False) returns
                User objects.
            to_dataframe: If `True`, returns `DataFrame`.
            **filters: Available filter parameters: ['id', 'name',
                'description', 'is_project_level_privilege']

        Examples:
            >>> Privilege.list_privileges(connection, to_dataframe=True,
            >>>                           is_project_level_privilege='True',
            >>>                           id=[1,2,3,4,5])
        """
        if to_dictionary and to_dataframe:
            raise ValueError(
                "Please select either `to_dictionary=True` or `to_dataframe=True`,"
                " but not both."
            )
        objects = helper.fetch_objects(
            connection=connection,
            api=security.get_privileges,
            limit=None,
            filters=filters,
        )
        if to_dictionary:
            return objects
        elif to_dataframe:
            return DataFrame(objects)
        else:
            return [cls.from_dict(source=obj, connection=connection) for obj in objects]

    def add_to_user(self, users: list[Union["User", str]]) -> None:
        """Add privilege to user.

        Args:
            users: list of `User` objects or names.
        """
        from mstrio.users_and_groups.user import User

        if isinstance(users, str):
            users = [User(self.connection, name=users)]
        elif isinstance(users, User):
            users = [users]
        elif hasattr(users, '__iter__') and all(isinstance(el, str) for el in users):
            users = [User(self.connection, name=user) for user in users]
        for user in users:
            user.grant_privilege(self.id)

    def revoke_from_user(self, users: list[Union["User", str]]) -> None:
        """Revoke privilege from user.

        Args:
            users: list of `User` objects or names.
        """
        from mstrio.users_and_groups.user import User

        if isinstance(users, str):
            users = [User(self.connection, name=users)]
        elif isinstance(users, User):
            users = [users]
        elif hasattr(users, '__iter__') and all(isinstance(el, str) for el in users):
            users = [User(self.connection, name=user) for user in users]
        for user in users:
            user.revoke_privilege(self.id)

    def add_to_user_group(self, groups: list[Union["UserGroup", str]]) -> None:
        """Add privilege to User Group.

        Args:
            groups: list of `UserGroup` objects or names.
        """
        if isinstance(groups, str):
            groups = [UserGroup(self.connection, name=groups)]
        elif isinstance(groups, UserGroup):
            groups = [groups]
        elif hasattr(groups, '__iter__') and all(isinstance(el, str) for el in groups):
            groups = [UserGroup(self.connection, name=group) for group in groups]
        for group in groups:
            group.grant_privilege(self.id)

    def revoke_from_user_group(self, groups: list[Union["UserGroup", str]]) -> None:
        """Revoke privilege from User Group.

        Args:
            groups: list of `UserGroup` objects or names.
        """
        if isinstance(groups, str):
            groups = [UserGroup(self.connection, name=groups)]
        elif isinstance(groups, UserGroup):
            groups = [groups]
        elif hasattr(groups, '__iter__') and all(isinstance(el, str) for el in groups):
            groups = [UserGroup(self.connection, name=group) for group in groups]
        for group in groups:
            group.revoke_privilege(self.id)

    @staticmethod
    def _validate_privileges(
        connection: Connection,
        privileges: 'Privilege | int | str | list[Privilege | int | str]',
    ) -> list[dict]:
        """This function validates if the privilege ID/Name/Object is valid and
        returns the IDs.

        If invalid, raise ValueError.
        """

        all_privileges = Privilege.list_privileges(
            connection=connection, to_dictionary=True
        )
        validated = []

        privileges = privileges if isinstance(privileges, list) else [privileges]

        for privilege in privileges:
            is_str_name = isinstance(privilege, str) and len(privilege) > 3
            is_str_id = isinstance(privilege, str) and 0 < len(privilege) <= 3
            is_int_id = isinstance(privilege, int) and 0 <= privilege < 300
            privilege_ok = False

            if is_str_name:
                temp_priv = helper.filter_list_of_dicts(all_privileges, name=privilege)
                privilege_ok = bool(temp_priv)
            elif is_str_id or is_int_id:
                temp_priv = helper.filter_list_of_dicts(
                    all_privileges, id=str(privilege)
                )
                privilege_ok = bool(temp_priv)

            if privilege_ok:
                validated.append(
                    {'id': temp_priv[0]['id'], 'name': temp_priv[0]['name']}
                )
            elif isinstance(privilege, Privilege):
                validated.append({'id': privilege.id, 'name': privilege.name})
            else:
                docs_url = (
                    "https://www2.microstrategy.com/producthelp/Current"
                    + "/ReferenceFiles/reference/com/microstrategy/webapi"
                    + "/EnumDSSXMLPrivilegeTypes.html"
                )
                msg = (
                    f"'{privilege}' is not a valid privilege. Possible values can be "
                    f"found in EnumDSSXMLPrivilegeTypes: \n" + docs_url
                )
                helper.exception_handler(msg, exception_type=ValueError)
        return validated

    @property
    def description(self):
        return self._description

    @property
    def categories(self):
        return self._categories

    @property
    def is_project_level_privilege(self):
        return self._is_project_level_privilege


class PrivilegeList:
    """Class for browsing MSTR privileges.

    Attributes:
        connection: A MicroStrategy connection object
    """

    def __init__(self, connection: Connection):
        """Initialize PrivilegeList object.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
        """
        self.connection = connection
        self.__privileges = Privilege.list_privileges(self.connection)

    @property
    def privileges(self):
        """Returns privileges list."""
        return self.__privileges

    def to_dict(self):
        """Returns list of privileges dicts."""
        return {int(p.id): p.name for p in self.__privileges}

    def to_dataframe(self):
        """Returns DataFrame with privileges."""
        return DataFrame(
            [[p.id, p.name, p.description, p.categories] for p in self.__privileges]
        )



import logging
from typing import Optional

from packaging import version

from mstrio import config
from mstrio.api import subscriptions as subscriptions_
from mstrio.connection import Connection
from mstrio.utils import helper
from mstrio.utils.version_helper import class_version_handler, method_version_handler

from . import (
    CacheUpdateSubscription,
    EmailSubscription,
    FileSubscription,
    FTPSubscription,
    HistoryListSubscription,
    Subscription,
)
from .content import Content
from .delivery import Delivery

logger = logging.getLogger(__name__)


@method_version_handler('11.2.0203')
def list_subscriptions(
    connection: Connection,
    project_id: str | None = None,
    project_name: str | None = None,
    to_dictionary: bool = False,
    limit: int | None = None,
    **filters,
) -> list["Subscription"] | list[dict]:
    """Get all subscriptions per project as list of Subscription objects or
    dictionaries.

    Optionally filter the subscriptions by specifying filters.
    Specify either `project_id` or `project_name`.
    When `project_id` is provided (not `None`), `project_name` is omitted.

    Note:
    When `project_id` is `None` and `project_name` is `None`,
    then its value is overwritten by `project_id` from `connection` object.

    Args:
        connection(object): MicroStrategy connection object
        project_id: Project ID
        project_name: Project name
        to_dictionary: If True returns a list of subscription dicts,
            otherwise (default) returns a list of subscription objects
        limit: limit the number of elements returned. If `None` (default), all
            objects are returned.
        **filters: Available filter parameters: ['id', 'multiple_contents',
            'name', 'editable', 'allow_delivery_changes'
            'allow_personalization_changes', 'allow_unsubscribe',
            'date_created', 'date_modified', 'owner', 'delivery']
    """
    project_id = helper.get_valid_project_id(
        connection=connection,
        project_id=project_id,
        project_name=project_name,
        with_fallback=not project_name,
    )
    chunk_size = (
        1000
        if version.parse(connection.iserver_version) >= version.parse('11.3.0300')
        else 1000000
    )
    msg = 'Error getting subscription list.'
    objects = helper.fetch_objects_async(
        connection=connection,
        api=subscriptions_.list_subscriptions,
        async_api=subscriptions_.list_subscriptions_async,
        limit=limit,
        chunk_size=chunk_size,
        filters=filters,
        error_msg=msg,
        dict_unpack_value="subscriptions",
        project_id=project_id,
    )

    if to_dictionary:
        return objects
    else:
        return [
            dispatch_from_dict(
                source=obj,
                connection=connection,
                project_id=project_id,
            )
            for obj in objects
        ]


DeliveryMode = Delivery.DeliveryMode
subscription_type_from_delivery_mode_dict = {
    DeliveryMode.CACHE: CacheUpdateSubscription,
    DeliveryMode.EMAIL: EmailSubscription,
    DeliveryMode.FILE: FileSubscription,
    DeliveryMode.FTP: FTPSubscription,
    DeliveryMode.HISTORY_LIST: HistoryListSubscription,
}


def get_subscription_type_from_delivery_mode(mode: DeliveryMode):
    """Returns the subscription type of the provided Delivery Mode.

    Args:
        mode: DeliveryMode object of which to get the subscription type"""
    return subscription_type_from_delivery_mode_dict.get(mode, Subscription)


def dispatch_from_dict(source: dict, connection: Connection, project_id: str):
    """Returns the subscription type object from the provided source

    Args:
        source: dictionary of an object to return from the specified
            subscription
        connection: MicroStrategy connection object returned
            by `connection.Connection()`
        project_id: Project ID"""
    delivery_mode = DeliveryMode(source["delivery"]["mode"])
    sub_type = get_subscription_type_from_delivery_mode(delivery_mode)
    return sub_type.from_dict(source, connection, project_id)


@class_version_handler('11.2.0203')
class SubscriptionManager:
    """Manage subscriptions."""

    def __init__(
        self,
        connection: Connection,
        project_id: str | None = None,
        project_name: str | None = None,
    ):
        """Initialize the SubscriptionManager object.
        Specify either `project_id` or `project_name`.
        When `project_id` is provided (not `None`), `project_name` is omitted.

        Args:
            connection: MicroStrategy connection object returned
                by `connection.Connection()`
            project_id: Project ID
            project_name: Project name
        """
        self.connection = connection
        self.project_id = helper.get_valid_project_id(
            connection=connection,
            project_id=project_id,
            project_name=project_name,
        )

    def list_subscriptions(
        self, to_dictionary: bool = False, limit: int | None = None, **filters
    ):
        """Get all subscriptions as list of Subscription objects or
        dictionaries.

        Optionally filter the subscriptions by specifying filters.

        Args:
            to_dictionary: If True returns a list of subscription dicts,
                otherwise returns a list of subscription objects
            limit: limit the number of elements returned. If `None` (default),
                all objects are returned.
            **filters: Available filter parameters: ['id', 'name', 'editable',
                'allowDeliveryChanges', 'allowPersonalizationChanges',
                'allowUnsubscribe', 'dateCreated', 'dateModified', 'owner',
                'schedules', 'contents', 'recipients', 'delivery']
        """
        return list_subscriptions(
            connection=self.connection,
            project_id=self.project_id,
            to_dictionary=to_dictionary,
            limit=limit,
            **filters,
        )

    def delete(
        self, subscriptions: list[Subscription] | list[str], force=False
    ) -> bool:
        """Deletes all passed subscriptions. Returns True if successfully
        removed all subscriptions.

        Args:
            subscriptions: list of subscriptions to be deleted
        """
        subscriptions = (
            subscriptions if isinstance(subscriptions, list) else [subscriptions]
        )
        if not subscriptions and config.verbose:
            logger.info('No subscriptions passed.')
        else:
            temp_subs = []
            for subscription in subscriptions:
                if not isinstance(subscription, Subscription):
                    subscription = Subscription(
                        connection=self.connection,
                        id=subscription,
                        project_id=self.project_id,
                    )
                temp_subs.append(subscription)
            subscriptions = temp_subs
            succeeded = 0
            user_input = 'N'
            if not force:
                to_be_deleted = [
                    f"Subscription '{sub.name}' with ID: '{sub.id}'"
                    for sub in subscriptions
                ]
                print("Found subscriptions:")
                for sub in to_be_deleted:
                    print(sub)
                user_input = input(
                    "Are you sure you want to delete all of them? [Y/N]: "
                )
            if force or user_input == 'Y':
                succeeded = 0
                for subscription in subscriptions:
                    response = subscriptions_.remove_subscription(
                        self.connection,
                        subscription.id,
                        self.project_id,
                        error_msg=(
                            f"Subscription '{subscription.name}' with id "
                            f"'{subscription.id}' could not be deleted."
                        ),
                        exception_type=UserWarning,
                    )
                    if response.ok:
                        succeeded += 1
                        if config.verbose:
                            logger.info(
                                f"Deleted subscription '{subscription.name}' "
                                f"with ID '{subscription.id}'."
                            )

                return succeeded == len(subscriptions)

    def execute(self, subscriptions: list[Subscription] | list[str]):
        """Executes all passed subscriptions.

        Args:
            subscriptions: list of subscriptions to be executed
        """
        if not subscriptions and config.verbose:
            logger.info('No subscriptions passed.')
        else:
            subscriptions = (
                subscriptions if isinstance(subscriptions, list) else [subscriptions]
            )
            for subscription in subscriptions:
                if not isinstance(subscription, Subscription):
                    subscription = Subscription(
                        connection=self.connection,
                        id=subscription,
                        project_id=self.project_id,
                    )
                if subscription.delivery.mode in (
                    'EMAIL',
                    'FILE',
                    'HISTORY_LIST',
                    'FTP',
                ):
                    subscription.execute()
                else:
                    msg = (
                        f"Subscription '{subscription.name}' with ID "
                        f"'{subscription.id}' could not be executed. Delivery mode "
                        f"'{subscription.delivery.mode}' is not supported."
                    )
                    helper.exception_handler(msg, UserWarning)

    @method_version_handler('11.3.0000')
    def available_bursting_attributes(self, content: dict | Content):
        """Get a list of available attributes for bursting feature, for a given
        content.

        Args:
            content: content dictionary or Content object
                (from subscription.content)
        """
        c_id = content['id'] if isinstance(content, dict) else content.id
        c_type = content['type'] if isinstance(content, dict) else content.type

        response = subscriptions_.bursting_attributes(
            self.connection, self.project_id, c_id, c_type.upper()
        )

        if response.ok:
            return response.json()['burstingAttributes']

    @method_version_handler('11.3.0000')
    def available_recipients(
        self,
        content_id: str | None = None,
        content_type: str | None = None,
        content: Optional["Content"] = None,
        delivery_type='EMAIL',
    ) -> list[dict]:
        """List available recipients for a subscription contents.
        Specify either both `content_id` and `content_type` or just `content`
        object.

        Args:
            content_id: ID of the content
            content_type: type of the content
            content: Content object
            delivery_type: The delivery of the subscription, available values
                are: [EMAIL, FILE, PRINTER, HISTORY_LIST, CACHE, MOBILE, FTP].
        """
        if content_id and content_type:
            pass
        elif isinstance(content, Content):
            content_id = content.id
            content_type = content.type
        else:
            helper.exception_handler(
                'Specify either a content ID and type or content object.', ValueError
            )

        body = {
            "contents": [{"id": content_id, "type": content_type}],
        }

        response = subscriptions_.available_recipients(
            self.connection, self.project_id, body, delivery_type
        )

        if response.ok and config.verbose:
            return response.json()['recipients']



from time import sleep
from mstrio.connection import get_connection
from mstrio.project_objects.datasets import list_all_cubes, OlapCube

# connect to environment without providing user credentials
# variable `workstationData` is stored within Workstation
connection = get_connection(workstationData, project_name='MicroStrategy Tutorial')

# List available Cubes (limited to 10)
cubes = list_all_cubes(connection, limit=10)
for cube in cubes:
    print(cube)

# Check available attributes and metrics of a Cube
sample_cube_id = '42FF415D4E162846C87D4FAD8B26BF4E'
sample_cube = OlapCube(connection, id=sample_cube_id)
attributes = sample_cube.attributes
metrics = sample_cube.metrics
print(f"Attributes: {attributes}\nMetrics: {metrics}")

# Choose attributes and metrics
call_center_attribute = attributes[0].get('id')
category_attribute = attributes[1].get('id')

profit_metric = metrics[0].get('id')
profit_margin_metric = metrics[1].get('id')

# Cube needs to be published before we can apply filters
sample_cube.publish()
sleep(5)

# Filter which attributes and metrics to use in a dataframe
sample_cube.apply_filters(
    attributes=[call_center_attribute, category_attribute],
    metrics=[profit_metric, profit_margin_metric],
)

# Create a dataframe from a Cube
dataframe = sample_cube.to_dataframe()
print(dataframe.head())



import logging
from collections import defaultdict
from enum import auto
from typing import TYPE_CHECKING

from mstrio import config
from mstrio.api import transmitters
from mstrio.types import ObjectTypes
from mstrio.users_and_groups import User
from mstrio.utils.entity import DeleteMixin, Entity
from mstrio.utils.enum_helper import AutoName, get_enum_val
from mstrio.utils.helper import (
    Dictable,
    fetch_objects,
    get_args_from_func,
    get_default_args_from_func,
)
from mstrio.utils.response_processors import objects as objects_processors
from mstrio.utils.translation_mixin import TranslationMixin
from mstrio.utils.version_helper import class_version_handler, method_version_handler

if TYPE_CHECKING:
    from mstrio.connection import Connection

logger = logging.getLogger(__name__)


class RecipientFieldType(AutoName):
    TO = auto()
    CC = auto()
    BCC = auto()


class TransmitterDeliveryType(AutoName):
    EMAIL = auto()
    FTP = auto()
    FILE = auto()
    PRINT = auto()
    IPHONE = auto()
    IPAD = auto()
    ANDROID = auto()
    UNSUPPORTED = auto()


class EmailTransmitterProperties(Dictable):
    """Representation of email transmitter properties object. Those properties
    are specific for Email Transmitters.

    Attributes:
        sender_display_name: name used in the email headers to indicate message
            sender
        sender_email_address: address used in the email headers to indicate
            message sender
        reply_to_display_name: name used in the email headers to indicate
            destination for replies
        reply_to_email_address: address used in the email headers to indicate
            destination for replies
        recipient_field_type: default field in which recipient address appear
            (eg. to/cc/bcc). Default value is `RecipientFieldType.TO`.
        save_message_to_file: specifies whether to save message output to file,
            default value is `False`
        send_message_via_smtp: specifies whether to send messages to recipients
            via SMTP, default value is `False`
        save_file_path: the folder path to save the email output,
            default is `None`
        notify_on_success: specifies whether to request notification on success,
            default is `False`
        notify_on_failure: specifies whether to request notification on failure,
            default is `False`
        notification_email_address: email address for notification, default is
            `None`
    """

    _FROM_DICT_MAP = {'recipient_field_type': RecipientFieldType}

    def __init__(
        self,
        sender_display_name: str,
        sender_email_address: str,
        reply_to_display_name: str,
        reply_to_email_address: str,
        recipient_field_type: RecipientFieldType = RecipientFieldType.TO,
        save_message_to_file: bool = False,
        send_message_to_file: bool = False,
        send_message_via_smtp: bool = False,
        save_file_path: str | None = None,
        notify_on_success: bool = False,
        notify_on_failure: bool = False,
        notification_email_address: str | None = None,
    ):
        self.sender_display_name = sender_display_name
        self.sender_email_address = sender_email_address
        self.reply_to_display_name = reply_to_display_name
        self.reply_to_email_address = reply_to_email_address
        self.recipient_field_type = recipient_field_type
        self.save_message_to_file = save_message_to_file
        self.send_message_to_file = send_message_to_file
        self.send_message_via_smtp = send_message_via_smtp
        self.save_file_path = save_file_path
        self.notify_on_success = notify_on_success
        self.notify_on_failure = notify_on_failure
        self.notification_email_address = notification_email_address


@method_version_handler('11.3.0100')
def list_transmitters(
    connection: "Connection",
    to_dictionary: bool = False,
    limit: int | None = None,
    **filters,
) -> list["Transmitter"] | list[dict]:
    """Get all transmitters as list of Transmitter objects or
    dictionaries.

    Optionally filter the transmitters by specifying filters.

    Args:
        connection(object): MicroStrategy connection object
        to_dictionary: If True returns a list of transmitter dicts,
            otherwise returns a list of transmitter objects
        limit: limit the number of elements returned. If `None` (default), all
            objects are returned.
        **filters: Available filter parameters: ['id', 'name', 'date_created',
            'date_modified', 'description', 'delivery_type',
             'email_transmitter_properties']
    """
    return Transmitter._list_transmitters(
        connection=connection, to_dictionary=to_dictionary, limit=limit, **filters
    )


@class_version_handler('11.3.0100')
class Transmitter(Entity, DeleteMixin, TranslationMixin):
    """Object representation of MicroStrategy Transmitter object

    Attributes:
        name: transmitter's name
        id: transmitter's id
        description: transmitter's description
        delivery_type: type of the transmitter
        email_transmitter_properties: properties specific to email transmitters
            (available only when `delivery_type` equals
            `TransmitterDeliveryType.EMAIL`)
    """

    _OBJECT_TYPE = ObjectTypes.SUBSCRIPTION_TRANSMITTER
    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        'owner': User.from_dict,
        'delivery_type': TransmitterDeliveryType,
        'email_transmitter_properties': EmailTransmitterProperties.from_dict,
    }
    _API_GETTERS = {
        (
            'abbreviation',
            'type',
            'subtype',
            'ext_type',
            'date_created',
            'date_modified',
            'version',
            'owner',
            'icon_path',
            'view_media',
            'ancestors',
            'certified_info',
            'acl',
            'comments',
        ): objects_processors.get_info,
        (
            'id',
            'name',
            'description',
            'delivery_type',
            'email_transmitter_properties',
        ): transmitters.get_transmitter,
    }
    _API_DELETE = staticmethod(transmitters.delete_transmitter)
    _API_PATCH: dict = {
        ("name", "description", "email_transmitter_properties"): (
            transmitters.update_transmitter,
            "put",
        ),
        'comments': (objects_processors.update, 'partial_put'),
    }
    _PATCH_PATH_TYPES = {
        "name": str,
        "description": str,
        "email_transmitter_properties": dict,
    }

    def __init__(
        self,
        connection: "Connection",
        id: str | None = None,
        name: str | None = None,
    ):
        """Initialize Transmitter object.

        Note:
            In case both `id` and `name` are provided, then `id` is used when
            fetching.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
            id (str, optional): Identifier of a pre-existing transmitter
            name (str, optional): Name of the pre-existing transmitter

        Raises:
            `ValueError` when neither `id` nor `name` are specified.
        """

        if id is None and name is None:
            raise ValueError(
                "Please specify either 'id' or 'name' parameter in the constructor."
            )

        if id is None:
            objects_info = Transmitter._list_transmitters(
                connection=connection,
                name=name,
                to_dictionary=True,
            )
            if objects_info:
                object_info, object_info["connection"] = objects_info[0], connection
                self._init_variables(**object_info)
            else:
                raise ValueError(f"There is no Transmitter: '{name}'")
        else:
            super().__init__(connection=connection, object_id=id)

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self._delivery_type = (
            TransmitterDeliveryType(kwargs["delivery_type"])
            if kwargs.get("delivery_type")
            else None
        )
        self.email_transmitter_properties = (
            EmailTransmitterProperties.from_dict(kwargs["email_transmitter_properties"])
            if kwargs.get("email_transmitter_properties")
            else None
        )

    @classmethod
    def create(
        cls,
        connection: "Connection",
        name: str,
        delivery_type: str | TransmitterDeliveryType,
        description: str | None = None,
        email_transmitter_properties: dict | EmailTransmitterProperties | None = None,
    ) -> "Transmitter":
        """Create transmitter.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
            delivery_type: type of the transmitter
            name (str): transmitter's name
            description (str): transmitter's description
            email_transmitter_properties (dict or object): properties specific
                to email transmitter. In transmitter with type `email` those
                properties must be provided. Otherwise they cannot be provided.

        Returns:
            `Transmitter` object
        """
        delivery_type = get_enum_val(delivery_type, TransmitterDeliveryType)
        email_transmitter_properties = (
            email_transmitter_properties.to_dict()
            if isinstance(email_transmitter_properties, EmailTransmitterProperties)
            else email_transmitter_properties
        )
        body = {
            'name': name,
            'description': description,
            'deliveryType': delivery_type,
            'emailTransmitterProperties': email_transmitter_properties,
        }

        res = transmitters.create_transmitter(connection, body).json()
        if config.verbose:
            logger.info(
                f"Successfully created transmitter named: '{res.get('name')}' "
                f"with ID: '{res.get('id')}'"
            )
        return cls.from_dict(res, connection)

    def alter(
        self,
        name: str | None = None,
        description: str | None = None,
        email_transmitter_properties: dict | EmailTransmitterProperties | None = None,
        comments: str | None = None,
    ):
        """Alter transmitter properties.

        Args:
            name (str): transmitter's name
            description (str): transmitter's description
            email_transmitter_properties (dict or object): properties specific
                to email transmitter. Only in transmitter with type `email`
                altering those properties is possible
            comments: long description of the transmitter
        """
        email_transmitter_properties = (
            email_transmitter_properties.to_dict()
            if isinstance(email_transmitter_properties, EmailTransmitterProperties)
            else email_transmitter_properties
        )
        func = self.alter
        args = get_args_from_func(func)
        defaults = get_default_args_from_func(func)
        defaults_dict = dict(zip(args[-len(defaults) :], defaults)) if defaults else {}
        local = locals()
        properties = defaultdict(dict)
        for property_key in defaults_dict.keys():
            if local[property_key] is not None:
                properties[property_key] = local[property_key]
        self._alter_properties(**properties)

    @classmethod
    def _list_transmitters(
        cls,
        connection: "Connection",
        to_dictionary: bool = False,
        limit: int | None = None,
        **filters,
    ) -> list["Transmitter"] | list[dict]:
        """Helper method for listing transmitters."""
        objects = fetch_objects(
            connection=connection,
            api=transmitters.get_transmitters,
            limit=limit,
            filters=filters,
            dict_unpack_value="transmitters",
        )

        if to_dictionary:
            return objects
        return [
            Transmitter.from_dict(source=obj, connection=connection) for obj in objects
        ]

    @property
    def delivery_type(self):
        return self._delivery_type



from mstrio.connection import get_connection
from mstrio.users_and_groups.user import list_users

# connect to environment without providing user credentials
# variable `workstationData` is stored within Workstation
conn = get_connection(workstationData, project_name='MicroStrategy Tutorial')

# change password for all users
for user in list_users(conn, name_begins='User_S'):
    user.alter(password='new_password')

# change user password
user = list_users(conn, name_begins='User_S')[0]
user.alter(password='new_password')

# change user password with option require_new_password
user.alter(password='new_password', require_new_password=True)

# change password for all users with option require_new_password
for user in list_users(conn, name_begins='User_S'):
    user.alter(password='new_password', require_new_password=True)



"""This is the demo script to show how to manage languages. This script will
not work without replacing parameters with real values. Its basic goal is to
present what can be done with this module and to ease its usage.
"""

from mstrio.connection import get_connection
from mstrio.server.language import list_interface_languages, list_languages, Language

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Project to connect to

# Create connection based on workstation data
CONN = get_connection(workstationData, project_name=PROJECT_NAME)

# List languages
language_list = list_languages(connection=CONN)

# List interface languages
interface_language_list = list_interface_languages(connection=CONN)

# Define variables which can be later used in a script
LANGUAGE_NAME = $language_name
LANGUAGE_ID = $language_id
LANGUAGE_NEW_NAME = $language_new_name
BASE_LANG = $base_lang
INTERFACE_LANG_ID = $interface_lang_id
MINUTES15 = $minutes15
MINUTES30 = $minutes30
HOUR = $hour
DAY = $day
HOUR_OF_DAY = $hour_of_day
WEEK = $week
MONTH = $month
QUARTER = $quarter
YEAR = $year

# Create a language
# For base_language you can either specify:
# - Language class object
# - specific Language ID
# - specific Language lcid
lang = Language.create(
    connection=CONN,
    name=LANGUAGE_NAME,
    base_language=BASE_LANG,
    interface_language_id=INTERFACE_LANG_ID,
    formatting_settings=Language.TimeInterval(
        minutes15=MINUTES15,
        minutes30=MINUTES30,
        hour=HOUR,
        day=DAY,
        hour_of_day=HOUR_OF_DAY,
        week=WEEK,
        month=MONTH,
        quarter=QUARTER,
        year=YEAR
    )
)

# Initialise a language
# By ID
lang2 = Language(connection=CONN, id=LANGUAGE_ID)
# By name
lang2 = Language(connection=CONN, name=LANGUAGE_NAME)

# Alter a language
lang.alter(
    name = LANGUAGE_NEW_NAME,
    formatting_settings=Language.TimeInterval(
        minutes15=MINUTES15,
        minutes30=MINUTES30,
        hour=HOUR,
        day=DAY,
        hour_of_day=HOUR_OF_DAY,
        week=WEEK,
        month=MONTH,
        quarter=QUARTER,
        year=YEAR
    )
)

# Delete a language
lang.delete(force=True)



import datetime as dt
from base64 import b64encode


class Encoder:
    """Internal method for converting a Pandas DataFrame to MicroStrategy
    compliant base64 encoded JSON.

    When creating a data set, MicroStrategy APIs require the tabular data to
    have been transformed first into JSON and then into a base 64 encoded string
    before it is transmitted to the Intelligence Server via the REST API to
    create the data set. This class uses Pandas to handle transforming
    the DataFrame into a JSON representation of the data. For single-table data
    sets, MicroStrategy APIs require the JSON data to be formatted using
    the 'records' orientation from Pandas. Conversely, multi-table data sets
    require the JSON data to have a 'values' orientation. Based on the data set
    type, the correct encoding strategy is applied and the data is then encoded.

    Attributes:
        __b64_data: DataFrame converted into a Base-64 encoded JSON string.
        __orientation: For single-table data sets, "single"; for multi-table
        data sets, "multi".
        __table_type_orient_map: Mapping used when converting DataFrame rows
            into proper JSON orientation needed for data uploads.
    """

    __b64_data = None
    __orientation = None
    __table_type_orient_map = {"single": "records", "multi": "values"}

    def __init__(self, data_frame, dataset_type):
        """Inits Encoder with given data_frame and type.

        Args:
            data_frame: Pandas DataFrame to be converted.
            dataset_type (str): Dataset type. One of `single` or `multi` to
                correspond with single-table or multi-table sources.
        """
        self.__data_frame = data_frame

        # Sets the proper orientation
        if dataset_type not in self.__table_type_orient_map:
            allowed_types = ', '.join(self.__table_type_orient_map)
            raise ValueError(f"Table type should be one of {allowed_types}")
        else:
            self.__orientation = self.__table_type_orient_map[dataset_type]

    @property
    def encode(self):
        """Encode data in base 64."""
        for col in self.__data_frame:
            if isinstance(self.__data_frame[col][0], dt.date):
                self.__data_frame[col] = self.__data_frame[col].astype('str')
        self.__b64_data = b64encode(
            self.__data_frame.to_json(
                orient=self.__orientation, date_format='iso'
            ).encode('utf-8')
        ).decode('utf-8')
        # return base 64 encoded data to calling environment
        return self.__b64_data



import itertools
import logging
from concurrent.futures import as_completed
from typing import TYPE_CHECKING, Optional

from mstrio.api import browsing, objects
from mstrio.connection import Connection
from mstrio.helpers import IServerError
from mstrio.object_management.search_enums import (
    CertifiedStatus,
    SearchDomain,
    SearchPattern,
    SearchResultsFormat,
)
from mstrio.server import Environment
from mstrio.server.project import Project
from mstrio.types import ObjectSubTypes, ObjectTypes
from mstrio.users_and_groups import User
from mstrio.utils.entity import CopyMixin, Entity, EntityBase, MoveMixin
from mstrio.utils.helper import (
    _prepare_objects,
    exception_handler,
    get_args_from_func,
    get_enum_val,
    get_objects_id,
    merge_id_and_type,
)
from mstrio.utils.response_processors import objects as objects_processors
from mstrio.utils.sessions import FuturesSessionWithRenewal
from mstrio.utils.translation_mixin import TranslationMixin
from mstrio.utils.version_helper import class_version_handler, method_version_handler

if TYPE_CHECKING:
    from mstrio.types import TypeOrSubtype

logger = logging.getLogger(__name__)


@class_version_handler('11.3.0100')
class SearchObject(Entity, CopyMixin, MoveMixin, TranslationMixin):
    """Search object describing criteria that specify a search for objects.

    Attributes:
        connection: A MicroStrategy connection object
        id: Object ID
        name: Object name
        description: Object description
        abbreviation: Object abbreviation
        type: Object type
        subtype: Object subtype
        ext_type: Object extended type
        date_created: Creation time, DateTime object
        date_modified: Last modification time, DateTime object
        version: Version ID
        owner: Owner ID and name
        icon_path: Object icon path
        view_media: View media settings
        ancestors: List of ancestor folders
        acg: Access rights (See EnumDSSXMLAccessRightFlags for possible values)
        acl: Object access control list
    """

    _OBJECT_TYPE = ObjectTypes.SEARCH
    _FROM_DICT_MAP = {**Entity._FROM_DICT_MAP, 'owner': User.from_dict}
    _API_PATCH: dict = {
        **Entity._API_PATCH,
        'folder_id': (objects_processors.update, 'partial_put'),
    }

    def __init__(self, connection: "Connection", id: str) -> None:
        """Initialize SearchObject object and synchronize with server.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
            id: ID of SearchObject
        """
        super().__init__(connection=connection, object_id=id)


@method_version_handler('11.3.0000')
def quick_search(
    connection: Connection,
    project: Project | str | None = None,
    name: str | None = None,
    root: str | None = None,
    pattern: SearchPattern | int = SearchPattern.CONTAINS,
    object_types: Optional["TypeOrSubtype"] = None,
    get_ancestors: bool = False,
    cross_cluster: bool = False,
    hidden: bool | None = None,
    certified_status: CertifiedStatus = CertifiedStatus.ALL,
    limit: int | None = None,
    offset: int | None = None,
    to_dictionary: bool = True,
):
    """Use the stored results of the Quick Search engine to return
     search results and display them as a list. The Quick Search
     engine periodically indexes the metadata and stores the results in memory,
     making Quick Search very fast but with results that may not be the
     most recent.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`
        project (string): `Project` object or ID
        name(string): Value the search pattern is set to, which will
            be applied to the names of object types being searched. For example,
            search for all report objects (type) whose name begins with
            (pattern) B (name).
        root(string, optional): Folder ID of the root folder where the search
            will be performed.
        pattern(integer or enum class object): Pattern to search for,
            such as Begin With or Exactly. Possible values are available in
            ENUM mstrio.object_management.SearchPattern.
            Default value is CONTAINS (4).
        object_types(enum class object or integer or list of enum class objects
            or integers): Type(s) of object(s) to be searched, such as
            Folder, Attribute or User. Possible values available in ENUMs
            mstrio.types.ObjectTypes and
            mstrio.types.ObjectSubTypes
        get_ancestors(bool): Specifies whether to return the list of ancestors
            for each object
        certified_status(CertifiedStatus): Defines a search criteria based
            on the certified status of the object, possible values available in
            CertifiedStatus enum
        cross_cluster(bool): Perform search in all unique projects across the
            cluster,this parameter only takes affect for I-Server with
            cluster nodes.
        hidden(bool): Filter the result based on the 'hidden' field of objects.
            If not passed, no filtering is applied.
        limit (int): limit the number of elements returned. If `None` (default),
            all objects are returned.
        offset (int): Starting point within the collection of returned
            results. Used to control paging behavior. Default is 0.
        to_dictionary (bool): If True returns dicts, by default
            (False) returns objects.

    Returns:
         list of objects or list of dictionaries
    """
    from mstrio.utils.object_mapping import map_objects_list

    project_id = get_objects_id(project, Project)
    if object_types and not isinstance(object_types, list):
        object_types = [get_enum_val(object_types, (ObjectTypes, ObjectSubTypes))]
    elif object_types:
        object_types = [
            get_enum_val(t, (ObjectTypes, ObjectSubTypes)) for t in object_types
        ]
    pattern = get_enum_val(pattern, SearchPattern)
    certified_status = get_enum_val(certified_status, CertifiedStatus)
    resp = browsing.get_quick_search_result(
        connection,
        project_id=project_id,
        name=name,
        pattern=pattern,
        object_types=object_types,
        root=root,
        get_ancestors=get_ancestors,
        hidden=hidden,
        cross_cluster=cross_cluster,
        limit=limit,
        certified_status=certified_status,
        offset=offset,
    )
    objects = resp.json()["result"]
    if to_dictionary:
        return objects
    return map_objects_list(connection, objects)


@method_version_handler('11.3.0100')
def quick_search_from_object(
    connection: Connection,
    project: Project | str,
    search_object: SearchObject | str,
    include_ancestors: bool = False,
    include_acl: bool = False,
    subtypes: ObjectSubTypes | list[ObjectSubTypes] | int | list[int] | None = None,
    limit: int | None = None,
    offset: int | None = None,
    to_dictionary: bool = True,
):
    """Perform a quick search based on a predefined Search Object.
    Use an existing search object for Quick Search engine to return
    search results and display them as a list.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`
        project (string): `Project` object or ID
        search_object(SearchObject): Search object ID to retrieve result from
        include_ancestors(bool): Specifies whether to return the list of
            ancestors for each object
        include_acl(bool): Specifies whether to return the list of ACL for each
            object
        subtypes(list, int): This parameter is used to filter the objects in the
            result to include only the ones whose “subtype” is included in
            the values of this parameter
        limit (int): limit the number of elements returned. If `None` (default),
            all objects are returned.
        offset (int): Starting point within the collection of returned
            results. Used to control paging behavior. Default is 0.
        to_dictionary (bool): If True returns dicts, by default
            (False) returns objects.

    Returns:
        list of objects or list of dictionaries
    """
    from mstrio.utils.object_mapping import map_objects_list

    project_id = get_objects_id(project, Project)
    search_object_id = get_objects_id(search_object, SearchObject)
    subtypes = get_enum_val(subtypes, ObjectSubTypes)
    resp = browsing.get_quick_search_result_from_object(
        connection,
        project_id,
        search_object_id,
        subtypes=subtypes,
        include_ancestors=include_ancestors,
        include_acl=include_acl,
        limit=limit,
        offset=offset,
    )
    objects = resp.json()["result"]
    if to_dictionary:
        return objects
    return map_objects_list(connection, objects)


@method_version_handler('11.3.0000')
def get_search_suggestions(
    connection: Connection,
    project: Project | str | None = None,
    key: str | None = None,
    max_results: int = 4,
    cross_cluster: bool | None = None,
) -> list[str]:
    """Request search suggestions from the server.

    Args:
        connection (object): MicroStrategy REST API connection object
        project (string, optional): `Project` object or ID
        key (string, optional): value the search pattern is set to, which will
            be applied to the names of suggestions being searched
        max_results (int, optional): maximum number of items returned
            for a single request. Used to control paging behavior.
            Default value is `-1` for no limit.
        cross_cluster (bool, optional): perform search in all unique projects
            across the cluster, this parameter only takes effect for I-Server
            with cluster nodes. Default value is `None`

    Returns:
        list of search suggestions
    """
    project_id = get_objects_id(project, Project)
    return (
        browsing.get_search_suggestions(
            connection=connection,
            project_id=project_id,
            key=key,
            count=max_results,
            is_cross_cluster=cross_cluster,
        )
        .json()
        .get('suggestions')
    )


@method_version_handler('11.3.0000')
def full_search(
    connection: Connection,
    project: Project | str | None = None,
    name: str | None = None,
    pattern: SearchPattern | int = SearchPattern.CONTAINS,
    domain: SearchDomain | int = SearchDomain.PROJECT,
    root: str | None = None,
    object_types: Optional['TypeOrSubtype'] = None,
    uses_object_id: EntityBase | str | None = None,
    uses_object_type: ObjectTypes | int | None = None,
    uses_recursive: bool = False,
    uses_one_of: bool = False,
    used_by_object_id: EntityBase | str | None = None,
    used_by_object_type: ObjectTypes | str | None = None,
    used_by_recursive: bool = False,
    used_by_one_of: bool = False,
    begin_modification_time: str | None = None,
    end_modification_time: str | None = None,
    results_format: SearchResultsFormat | str = SearchResultsFormat.LIST,
    limit: int | None = None,
    offset: int | None = None,
    to_dictionary: bool = True,
    **filters,
) -> list[dict] | list[Entity]:
    """Perform a full metadata search and return results.

    Note:
        If you have a large number of objects in your environment, try to use
        `limit` and `offset` parameters to retrieve the results in batches.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`
        project (string): `Project` object or ID
        name(string): Value the search pattern is set to, which will
            be applied to the names of object types being searched. For example,
            search for all report objects (type) whose name begins with
            pattern) B (name).
        pattern(integer or enum class object): Pattern to search for,
            such as Begin With or Exactly. Possible values are available in
            ENUM mstrio.object_management.SearchPattern.
            Default value is CONTAINS (4).
        domain(integer or enum class object): Domain where the search
            will be performed, such as Local or Project. Possible values are
            available in ENUM mstrio.object_management.SearchDomain.
            Default value is PROJECT (2).
        root(string, optional): Folder ID of the root folder where the search
            will be performed.
        object_types(enum class object or integer or list of enum class objects
            or integers): Type(s) of object(s) to be searched, such as
            Folder, Attribute or User. Possible values available in ENUMs
            mstrio.types.ObjectTypes and
            mstrio.types.ObjectSubTypes
        uses_object_id(string): Constrain the search to only return
            objects which use the given object.
        uses_object_type(string): Constrain the search to only return
            objects which use the given object. Possible values
            available in ENUMs mstrio.types.ObjectTypes
        uses_recursive(boolean): Control the Intelligence server to
            also find objects that use the given objects indirectly. Default
            value is false.
        uses_one_of(boolean): Control the Intelligence server to also
            find objects that use one of or all of given objects indirectly.
            Default value is false.
        used_by_object_id(string): Constrain the search to only return
            objects which are used by the given object.
        used_by_object_type(string): Constrain the search to only return
            objects which are used by the given object. Possible values
            available in ENUM mstrio.types.ObjectTypes
        used_by_recursive(boolean, optional): Control the Intelligence server
            to also find objects that are used by the given objects indirectly.
            Default value is false.
        used_by_one_of(boolean): Control the Intelligence server to also
            find objects that are used by one of or all of given objects
            indirectly. Default value is false.
        begin_modification_time(string, optional): Field to filter request
            to return records newer than a given date in
            format 'yyyy-MM-dd'T'HH:mm:ssZ', for example 2021-04-04T06:33:32Z.
        end_modification_time(string, optional): Field to filter request
            to return records older than a given date in
            format 'yyyy-MM-dd'T'HH:mm:ssZ', for example 2022-04-04T06:33:32Z.
        results_format(SearchResultsFormat): either a list or a tree format
        to_dictionary (bool): If False returns objects, by default
            (True) returns dictionaries.
        limit (int): limit the number of elements returned. If `None` (default),
            all objects are returned.
        offset (int): Starting point within the collection of returned
            results. Used to control paging behavior. Default is 0.
        **filters: Available filter parameters: ['id', 'name', 'description',
            'date_created', 'date_modified', 'acg']

    Returns:
        list of objects or list of dictionaries
    """
    passed_params = locals()
    start_search_args = get_args_from_func(start_full_search)
    search_result_args = get_args_from_func(get_search_results)
    start_search_params = {
        k: v for k, v in passed_params.items() if k in start_search_args
    }
    search_result_params = {
        k: v for k, v in passed_params.items() if k in search_result_args
    }
    resp = start_full_search(**start_search_params)
    search_result_params["search_id"] = resp["id"]
    del search_result_params["filters"]
    search_result_params.update(**filters)
    return get_search_results(**search_result_params)


@method_version_handler('11.3.0000')
def start_full_search(
    connection: Connection,
    project: Project | str | None = None,
    name: str | None = None,
    object_types: Optional["TypeOrSubtype"] = None,
    pattern: SearchPattern | int = SearchPattern.CONTAINS,
    domain: SearchDomain | int = SearchDomain.PROJECT,
    root: str | None = None,
    uses_object_id: EntityBase | str | None = None,
    uses_object_type: ObjectTypes | ObjectSubTypes | int | None = None,
    uses_recursive: bool = False,
    uses_one_of: bool = False,
    used_by_object_id: EntityBase | str | None = None,
    used_by_object_type: ObjectTypes | ObjectSubTypes | int | None = None,
    used_by_recursive: bool = False,
    used_by_one_of: bool = False,
    begin_modification_time: str | None = None,
    end_modification_time: str | None = None,
) -> dict:
    """Search the metadata for objects in a specific project that
    match specific search criteria, and save the results in IServer memory.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`
        project (string): `Project` object or ID
        object_types(enum class object or integer or list of enum class objects
            or integers): Type(s) of object(s) to be searched, such as
            Folder, Attribute or User. Possible values available in ENUMs
            mstrio.types.ObjectTypes and
            mstrio.types.ObjectSubTypes
        name(string): Value the search pattern is set to, which will
            be applied to the names of object types being searched. For example,
            search for all report objects (type) whose name begins with
            (pattern) B (name).
        pattern(integer or enum class object): Pattern to search for,
            such as Begin With or Exactly. Possible values are available in
            ENUM mstrio.object_management.SearchPattern.
            Default value is CONTAINS (4).
        domain(integer or enum class object): Domain where the search
            will be performed, such as Local or Project. Possible values are
            available in ENUM mstrio.object_management.SearchDomain.
            Default value is PROJECT (2).
        root(string, optional): Folder ID of the root folder where the search
            will be performed.
        uses_object_id(string): Constrain the search to only return
            objects which use the given object.
        uses_object_type(int, ObjectTypes): Constrain the search to only return
            objects which use the given object. Possible values
            available in ENUMs mstrio.types.ObjectTypes
        uses_recursive(boolean): Control the Intelligence server to
            also find objects that use the given objects indirectly. Default
            value is false.
        uses_one_of(boolean): Control the Intelligence server to also find
             objects that use one of or all of given objects indirectly.
             Default value is false.
        used_by_object_id(string): Constrain the search to only return
            objects which are used by the given object.
        used_by_object_type(int, ObjectTypes): Constrain the search to only
            return objects which are used by the given object. Possible values
            available in ENUM mstrio.types.ObjectTypes
        used_by_recursive(boolean, optional): Control the Intelligence server
            to also find objects that are used by the given objects indirectly.
            Default value is false.
        used_by_one_of(boolean): Control the Intelligence server to also
            find objects that are used by one or all given objects
            indirectly. Default value is false.
        begin_modification_time(string, optional): Field to filter request
            to return records newer than a given date in
            format 'yyyy-MM-dd'T'HH:mm:ssZ', for example 2021-04-04T06:33:32Z.
        end_modification_time(string, optional): Field to filter request
            to return records older than a given date in
            format 'yyyy-MM-dd'T'HH:mm:ssZ', for example 2022-04-04T06:33:32Z.

    Returns:
        dictionary consisting of id (Search ID) and total items number
    """

    if uses_object_id and used_by_object_id:
        exception_handler(
            msg="It is not allowed to use both ‘uses_object’ and ‘used_by_object’ in "
            "one request.",
            exception_type=AttributeError,
        )

    uses_object = (
        merge_id_and_type(
            uses_object_id,
            uses_object_type,
            "Please provide both `uses_object_id` and `uses_object_type`.",
        )
        if uses_object_id
        else None
    )
    used_by_object = (
        merge_id_and_type(
            used_by_object_id,
            used_by_object_type,
            "Please provide both `used_by_object_id` and `uses_object_type`.",
        )
        if used_by_object_id
        else None
    )

    project_id = get_objects_id(project, Project)
    if object_types and not isinstance(object_types, list):
        object_types = [get_enum_val(object_types, (ObjectTypes, ObjectSubTypes))]
    elif object_types:
        object_types = [
            get_enum_val(t, (ObjectTypes, ObjectSubTypes)) for t in object_types
        ]
    pattern = get_enum_val(pattern, SearchPattern)
    domain = get_enum_val(domain, SearchDomain)
    resp = browsing.store_search_instance(
        connection=connection,
        project_id=project_id,
        name=name,
        pattern=pattern,
        domain=domain,
        root=root,
        object_types=object_types,
        uses_object=uses_object,
        uses_recursive=uses_recursive,
        uses_one_of=uses_one_of,
        used_by_object=used_by_object,
        used_by_recursive=used_by_recursive,
        used_by_one_of=used_by_one_of,
        begin_modification_time=begin_modification_time,
        end_modification_time=end_modification_time,
    )
    return resp.json()


@method_version_handler('11.3.0000')
def get_search_results(
    connection: Connection,
    search_id: str,
    project: Project | str | None = None,
    results_format: SearchResultsFormat = SearchResultsFormat.LIST,
    limit: int | None = None,
    offset: int | None = None,
    to_dictionary: bool = False,
    **filters,
):
    """Retrieve the results of a full metadata search previously stored in
    an instance in IServer memory, may be obtained by `start_full_search`.

    Note:
        If you have a large number of objects in your environment, try to use
        `limit` and `offset` parameters to retrieve the results in batches.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`
        search_id (str): Search ID (identifies the results of a previous search
            stored in IServer memory)
        project (string): `Project` object or ID
        results_format(SearchResultsFormat): either a list or a tree format
        to_dictionary (bool): If True returns dicts, by default
            (False) returns objects.
        limit (int): limit the number of elements returned. If `None` (default),
            all objects are returned.
        offset (int): Starting point within the collection of returned
            results. Used to control paging behavior. Default is 0.
        **filters: Available filter parameters: ['id', 'name', 'description',
            'date_created', 'date_modified', 'acg']

    Returns:
        list of objects or list of dictionaries
    """
    passed_params = locals()
    get_result_args = get_args_from_func(_get_search_result_tree_format)
    get_result_params = {k: v for k, v in passed_params.items() if k in get_result_args}
    project_id = get_objects_id(project, Project)
    get_result_params['project_id'] = project_id
    if results_format == SearchResultsFormat.TREE:
        logger.info(
            'Notice. When results_format is equal to TREE, results are always returned'
            ' in a dictionary format.'
        )
        return _get_search_result_tree_format(**get_result_params)
    get_result_params.update({**filters, 'to_dictionary': to_dictionary})
    return _get_search_result_list_format(**get_result_params)


def _get_search_result_list_format(
    connection: Connection,
    search_id: str,
    project_id: str | None = None,
    limit: int | None = None,
    offset: int | None = None,
    to_dictionary: bool = True,
    **filters,
):
    from mstrio.utils.object_mapping import map_objects_list

    try:
        response = browsing.get_search_results(
            connection=connection,
            search_id=search_id,
            project_id=project_id,
            limit=limit,
            offset=offset,
        )
    except IServerError as e:
        if 'Java heap space' in str(e):
            logger.warning(
                "If you have a large number of objects in your environment, "
                "try to use `limit` and `offset` parameters "
                "to retrieve the results in batches."
            )
        raise e
    objects = _prepare_objects(response.json(), filters)

    if to_dictionary:
        return objects
    return map_objects_list(connection, objects)


def _get_search_result_tree_format(
    connection: Connection,
    search_id: str,
    project_id: str | None = None,
    limit: int | None = None,
    offset: int | None = None,
):
    try:
        resp = browsing.get_search_results_tree_format(
            connection=connection,
            search_id=search_id,
            project_id=project_id,
            limit=limit,
            offset=offset,
        )
    except IServerError as e:
        if 'Java heap space' in str(e):
            logger.warning(
                "If you have a large number of objects in your environment, "
                "try to use `limit` and `offset` parameters "
                "to retrieve the results in batches."
            )
        raise e
    if resp.content:
        return resp.json()
    return {}


def find_objects_with_id(
    connection: Connection,
    object_id: str,
    projects: list[Project] | list[str] | None = None,
    to_dictionary=False,
) -> list[dict[str, dict | type[Entity]]]:
    """Find object by its ID only, without knowing project ID and object type.
    The search is performed by iterating over projects and trying to retrieve
    the objects with different type.

    Limitation:
        - Only object types supported by mstrio, i.g. present in
            `mstrio.types.ObjecTypes` enum, are used.
        - Configuration object are also not searched.

    Args:
        connection (Connection): MicroStrategy connection object returned by
            `connection.Connection()`.
        object_id (str): ID of an object.
        projects (list[Project] | list[str], optional): List of projects where
            to perform the search. By default, if no project are provides, the
            search is performed on all loaded projects.
        to_dictionary (bool, optional): If True returns dicts, by default
            (False) returns objects.

    Returns:
        Returns list of dict with the following structure:
        {
            'project_id': <str>,
            'object_data': <dict or object>
        }
    """
    env = Environment(connection)
    projects = projects if projects else env.list_loaded_projects()

    with FuturesSessionWithRenewal(connection=connection) as session:
        futures = []

        for project, obj_type in itertools.product(projects, ObjectTypes):
            project_id = project.id if isinstance(project, Project) else project

            future = objects.get_object_info_async(
                future_session=session,
                id=object_id,
                object_type=obj_type.value,
                project_id=project_id,
            )
            future.project_id = project_id
            futures.append(future)

    from mstrio.utils.object_mapping import map_object

    return [
        {
            'project_id': future.project_id,
            'object_data': (
                result.json()
                if to_dictionary
                else map_object(connection, result.json())
            ),
        }
        for future in as_completed(futures)
        if (result := future.result()) and result.ok
    ]



"""This is the demo script to show how to manage metrics. This script will not
work without replacing parameters with real values. Its basic goal is to
present what can be done with this module and to ease its usage.
"""

from mstrio.connection import get_connection
from mstrio.modeling import (
    DataType,
    DefaultSubtotals,
    Dimensionality,
    DimensionalityUnit,
    FormatProperty,
    list_metrics,
    Metric,
    MetricFormat,
    ObjectSubType,
    SchemaManagement,
    SchemaObjectReference,
    SchemaUpdateType
)
from mstrio.modeling.expression import Expression, Token

# For every object we want to reference using a SchemaObjectReference we need
# to provide an Object ID for. For the script to work correctly all occurences
# of `'<object_id>'` and others with form `<some_name>` need to be replaced with
# data specific to the object used.

# Define variables which can be later used in a script
PROJECT_NAME = $project_name  # Insert name of project here
FOLDER_ID = $folder_id  # Insert folder ID here
HIDDEN = $hidden  # Insert value for hidden attribute here

conn = get_connection(workstationData, PROJECT_NAME)

# Example metric data.
# Parts of this dictionary will be used in the later parts of this demo script
METRIC_DATA = {
    'name': '<metric_name>',
    'sub_type': ObjectSubType.METRIC,
    'destination_folder': FOLDER_ID,
    'expression': Expression(
        text='Sum(Cost)',
        tokens=[
            Token(
                value='<token_value>',
                type=Token.Type.FUNCTION,
                target=SchemaObjectReference(sub_type=ObjectSubType.FUNCTION, object_id='<object_id>')
            ),
            Token(value='(', type=Token.Type.CHARACTER),
            Token(
                value='<token_value>',
                type=Token.Type.OBJECT_REFERENCE,
                target=SchemaObjectReference(sub_type=ObjectSubType.FACT, object_id='<object_id>')
            ),
            Token(value=')', type=Token.Type.CHARACTER),
            Token(value='', type=Token.Type.END_OF_TEXT)
        ]
    ),
    'dimensionality': Dimensionality(
        dimensionality_units=[
            DimensionalityUnit(
                dimensionality_unit_type=DimensionalityUnit.DimensionalityUnitType.ATTRIBUTE,
                aggregation=DimensionalityUnit.Aggregation.LAST_IN_RELATIONSHIP,
                filtering=DimensionalityUnit.Filtering.APPLY,
                group_by=True,
                relative_position=None,
                target=SchemaObjectReference(
                    sub_type=ObjectSubType.ATTRIBUTE,
                    name='<attribute_name>',
                    object_id='<object_id>'
                ),
                axis_collection=None
            ),
            DimensionalityUnit(
                dimensionality_unit_type=DimensionalityUnit.DimensionalityUnitType.ATTRIBUTE,
                aggregation=DimensionalityUnit.Aggregation.FIRST_IN_RELATIONSHIP,
                filtering=DimensionalityUnit.Filtering.IGNORE,
                group_by=True,
                relative_position=None,
                target=SchemaObjectReference(
                    sub_type=ObjectSubType.ATTRIBUTE, name='<attribute_name>', object_id='<object_id>'
                ),
                axis_collection=None
            )
        ]
    ),
    'conditionality': Metric.Conditionality(
        filter=SchemaObjectReference(
            sub_type=ObjectSubType.FILTER, name='<filter_name>', object_id='<object_id>'
        ),
        embed_method=Metric.Conditionality.EmbedMethod.METRIC_INTO_REPORT_FILTER,
        remove_elements=True
    ),
    'metric_format_type': Metric.MetricFormatType.HTML_TAG,
    'metric_subtotals': [
        Metric.MetricSubtotal(
            SchemaObjectReference(ObjectSubType.METRIC_SUBTOTAL, object_id='<object_id>')
        )
    ],
    'aggregate_from_base': False,
    'formula_join_type': Metric.FormulaJoinType.INNER,
    'smart_total': Metric.SmartTotal.DECOMPOSABLE_FALSE,
    'data_type': DataType(DataType.Type.INTEGER, 10, 0),
    'format': MetricFormat(
        values=[FormatProperty(type=FormatProperty.FormatType.NUMBER_CATEGORY, value=0)],
        header=[
            FormatProperty(type=FormatProperty.FormatType.NUMBER_CATEGORY, value=4),
            FormatProperty(type=FormatProperty.FormatType.NUMBER_DECIMAL_PLACES, value=1),
        ],
    ),
    'subtotal_from_base': False,
    'column_name_alias': '<column_name_alias>',
    'thresholds': None,
    'hidden': HIDDEN,
}

# Metric management
# Get list of metrics, with examples of different conditions
list_of_all_metrics = list_metrics(connection=conn)
list_of_metrics_with_limit = list_metrics(connection=conn, limit=10)

# Define variables which can be later used in a script
METRIC_NAME = $metric_name  # Insert name of edited metric here
METRIC_ID = $metric_id  # Insert ID of edited metric here

list_of_metrics_by_name = list_metrics(connection=conn, name=METRIC_NAME)
list_of_metrics_by_id = list_metrics(connection=conn, id=METRIC_ID)
list_of_metrics_in_folder = list_metrics(connection=conn, root=FOLDER_ID)

# Get specific metric by id or name
metr = Metric(connection=conn, id=METRIC_ID)
metr = Metric(connection=conn, name=METRIC_NAME)

# Listing properties
properties = metr.list_properties()

# Create metric
metr = Metric.create(
    connection=conn,
    name=METRIC_DATA['name'],
    sub_type=METRIC_DATA['sub_type'],
    destination_folder=METRIC_DATA['destination_folder'],
    expression=METRIC_DATA['expression'],
    metric_format_type=METRIC_DATA['metric_format_type'],
    metric_subtotals=METRIC_DATA['metric_subtotals'],
    format=METRIC_DATA['format'],
)

# or do it with one-liner unpacking `METRIC_DATA` dictionary
# (delete metric first, as it would have the same name and cause error)
metr.delete(True)
metr = Metric.create(connection=conn, **METRIC_DATA)

# Any changes to a schema objects must be followed by schema_reload
# in order to use them in reports, dashboards and so on
schema_manager = SchemaManagement(connection=conn, project_id=conn.project_id)
task = schema_manager.reload(update_types=[SchemaUpdateType.LOGICAL_SIZE])

# MetricSubtotals
# During creation of `Metric` you can either use custom subtotals:
custom_subtotal = Metric.MetricSubtotal(
    definition=SchemaObjectReference(ObjectSubType.METRIC_SUBTOTAL, object_id='<object_id>')
)
# or default subtotals using `DefaultSubtotals`.:
default_subtotal = Metric.MetricSubtotal(definition=DefaultSubtotals.AVERAGE)

# Subtotals are also used to set the Dynamic Aggregation function and to set
# the Function for default subtotal (Total).

# To set the Dynamic Aggregation function we need to include a special subtotal
# in the metric_subtotals list. The definition has to be the default Aggregation
# subtotal, while the implementation needs to be set to the subtotal we want to
# set as the default dynamic aggregation function.
dynamic_aggregation = Metric.MetricSubtotal(
    definition=DefaultSubtotals.AGGREGATION, implementation=DefaultSubtotals.MODE
)

# To set the Function for default subtotal (Total) we do a similar approach, but
# instead of setting the definition to Aggregation, we set it to Total.
total_aggregation = Metric.MetricSubtotal(
    definition=DefaultSubtotals.TOTAL, implementation=DefaultSubtotals.MAXIMUM
)

# Now if we want to change the metrics subtotals we need to send the list of
# the totals we prepared as an alteration to metric_subtotals:
metr.alter(metric_subtotals=[default_subtotal, dynamic_aggregation, total_aggregation])

# Metric Format
# During creation of `Metric` and providing format in `MetricFormat`
# representation, `FormatProperty` class have to be used.
# The list of values with explanations for each Format Type is available in the
# docstrings of the class
new_format = MetricFormat(
    values=[FormatProperty(type=FormatProperty.FormatType.NUMBER_CATEGORY, value=0)],
    header=[
        FormatProperty(type=FormatProperty.FormatType.NUMBER_CATEGORY, value=4),
        FormatProperty(type=FormatProperty.FormatType.NUMBER_DECIMAL_PLACES, value=1),
    ],
)

# Setting every formatting available in Workstation is possible through Python.
# Below you can see a snippet on how to change the currency symbol to Euro and
# the currency symbol position to behind the number (10€).
currency_symbol_format = FormatProperty(
    type=FormatProperty.FormatType.NUMBER_CURRENCY_SYMBOL, value='€'
)
currency_symbol_position = FormatProperty(
    type=FormatProperty.FormatType.NUMBER_CURRENCY_POSITION, value='1'
)

new_format.values.append(currency_symbol_format)
new_format.values.append(currency_symbol_position)

# Depending on the `FormatProperty`s type, the value will be of different
# type (str, int, etc.). If it is related to color, use FormatProperty.Color
# with one of the three types of initialization: RGB, hex value or
# server readable form
rgb_color = FormatProperty(
    type=FormatProperty.FormatType.FONT_COLOR,
    value=FormatProperty.Color(red=255, green=102, blue=51)
)
hex_color = FormatProperty(
    type=FormatProperty.FormatType.FONT_COLOR, value=FormatProperty.Color(hex_value='#ff6633')
)
server_color = FormatProperty(
    type=FormatProperty.FormatType.FONT_COLOR, value=FormatProperty.Color(server_value='16737843')
)

new_format.values.append(rgb_color)

# Define variables which can be later used in a script
METRIC_NEW_NAME = $metric_new_name  # Insert new name of edited metric here
METRIC_NEW_DESCRIPTION = $metric_new_description  # Insert new description of edited metric here

# Alter metric using newly defined format
metr.alter(name=METRIC_NEW_NAME, description=METRIC_NEW_DESCRIPTION, format=new_format)

# Altering the Dimensionality and Conditionality of the Metric

# Altering Dimensionality and Conditionality can be done in two ways. They can
# be altered either through altering the Expression or their respective class,
# which is only possible when the metric is simple, as for compound metrics,
# both the classes are disabled and all the changes on them can only be done
# through altering the Expression.
# When both ways are used at once, the content of the Dimensionality and
# Conditionality classes has priority over Expression.

# Important thing to remember is when altering the Expression with new content,
# if the Metric already has a defined Dimensionality, Conditionality or both,
# they need to be included in the Expression tokens or they will be erased
# while altering the Expression.
# Alternatively when altering the Expression, providing the Dimensionality,
# Conditionality or both of them into the alter function will also preserve
# their respective contents in the metric without the necessity of including
# them into the contents of the Expression.

# Let us say that we want to change the Dimensionality to only work for the
# Item attribute, with the rules to ignore the report filter and to aggregate
# based on the first value of the lookup table and let us also say we want to
# change the Conditionality to be based around the East Region filter, with
# merging of the metric filter into the report filter and ignoring related report
# filter elements. The Dimensionality and Conditionality classes would then
# look like this:

dimensionality_example = Dimensionality(
    dimensionality_units=[
        DimensionalityUnit(
            dimensionality_unit_type=DimensionalityUnit.DimensionalityUnitType.ATTRIBUTE,
            aggregation=DimensionalityUnit.Aggregation.FIRST_IN_RELATIONSHIP,
            filtering=DimensionalityUnit.Filtering.IGNORE,
            group_by=True,
            relative_position=None,
            target=SchemaObjectReference(
                sub_type=ObjectSubType.ATTRIBUTE, name='<attribute_name>', object_id='<object_id>'
            ),
            axis_collection=None
        )
    ]
)

conditionality_example = Metric.Conditionality(
    filter=SchemaObjectReference(
        sub_type=ObjectSubType.FILTER, name='<filter_name>', object_id='<object_id>'
    ),
    embed_method=Metric.Conditionality.EmbedMethod.METRIC_INTO_REPORT_FILTER,
    remove_elements=True
)

# To achieve the same result with Expression, it would need to look like this:
expression_example = Expression(
    text='Sum(Cost)',
    tokens=[
        Token(
            value='<token_value>',
            type=Token.Type.FUNCTION,
            target=SchemaObjectReference(sub_type=ObjectSubType.FUNCTION, object_id='<object_id>')
        ),
        Token(value='(', type=Token.Type.CHARACTER),
        Token(
            value='<token_value>',
            type=Token.Type.OBJECT_REFERENCE,
            target=SchemaObjectReference(sub_type=ObjectSubType.FACT, object_id='<object_id>')
        ),
        Token(value=')', type=Token.Type.CHARACTER),
        # Here the dimensionality part begins:
        # Changes to dimensionality in the Expression need to be in {} brackets
        Token(value='{', type=Token.Type.CHARACTER),
        # The < symbolises aggregation by first value of the lookup table
        Token(value='<', type=Token.Type.CHARACTER),
        Token(
            value='<token_value>',
            type=Token.Type.OBJECT_REFERENCE,
            target=SchemaObjectReference(sub_type=ObjectSubType.ATTRIBUTE, object_id='<object_id>')
        ),
        # The % symbolises ignoring the report filter
        Token(value='%', type=Token.Type.CHARACTER),
        Token(value='}', type=Token.Type.CHARACTER),
        # Here the conditionality part begins:
        # Changes to conditionality in the Expression need to be in <> brackets
        Token(value='<', type=Token.Type.CHARACTER),
        Token(
            value='<token_value>',
            type=Token.Type.OBJECT_REFERENCE,
            target=SchemaObjectReference(sub_type=ObjectSubType.FILTER, object_id='<object_id>')
        ),
        # The @3 symbols represent merging the metric filter into the report filter
        Token(value=';', type=Token.Type.CHARACTER),
        Token(value='@', type=Token.Type.CHARACTER),
        Token(value='3', type=Token.Type.CHARACTER),
        # The ;- symbols represent setting the Ignore related filter elements to True
        Token(value=';', type=Token.Type.CHARACTER),
        Token(value='-', type=Token.Type.CHARACTER),
        Token(value='>', type=Token.Type.CHARACTER),
        Token(value='', type=Token.Type.END_OF_TEXT)
    ]
)

# Apply the change
metr.alter(expression=expression_example)

# Deleting metrics
metr.delete(force=True)



"""Add email address with a form `{username}@microstrategy.com` to every user
which is enabled but doesn't have an email address. For each successfully added
email address a message will be printed.
"""

from typing import List

from mstrio.connection import Connection, get_connection
from mstrio.users_and_groups import list_users, User


def add_email_to_new_users(
    connection: "Connection", domain="microstrategy.com"
) -> List["User"]:
    """Add email address with a form `{username}@microstrategy.com`
    to every user which is enabled but doesn't have an email address.
    For each successfully added email address a message will be printed.

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`
        domain: name of the domain in the email address (it should be
            provided without '@' symbol). Default value is "microstrategy.com".

    Returns:
        list of users to which email addresses where added
    """
    # get all users that are enabled
    all_users = list_users(connection=connection)
    users_ = [u for u in all_users if u.enabled]
    modified_users_ = []
    for user_ in users_:
        # add email address only for those users which don't have one
        if not user_.addresses:
            email_address = user_.username + '@' + domain
            user_.add_address(name=user_.username, address=email_address)
            modified_users_.append(user_)

    return modified_users_


# connect to environment without providing user credentials
# variable `workstationData` is stored within Workstation
conn = get_connection(workstationData, 'MicroStrategy Tutorial')

# execute adding email to new users
# if needed provide some domain - default is 'microstrategy.com'
add_email_to_new_users(conn)



from mstrio.connection import get_connection
from mstrio.datasources import list_datasource_connections
from mstrio.server import Project, list_languages
from mstrio.server.project_languages import DataLocalizationLanguage, SimpleLanguage

PROJECT_NAME = 'MicroStrategy Tutorial'  # Project to connect to

# Create connection based on workstation data
conn = get_connection(workstationData, project_name=PROJECT_NAME)

# Get project with a given name
project = Project(connection=conn, name=PROJECT_NAME)

# Get list of all available datasource connections
datasource_connections = list_datasource_connections(conn)
# Get a datasource connection with a given name to use it for the data language
xquery_datasource_connection = list_datasource_connections(conn, name='XQuery(1)')[0]

# Get a list of all available languages
languages = list_languages(connection=conn)
# Get languages needed to be added to a project
polish_language = list_languages(connection=conn, name='Polish')[0]
danish_language = list_languages(connection=conn, name='Danish (Denmark)')[0]
dutch_language = list_languages(connection=conn, name='Dutch (Netherlands)')[0]

# Convert Polish and Danish languages to DataLocalizationLanguage objects
polish_data_language = DataLocalizationLanguage(
    id=polish_language.id,
    name=polish_language.name,
    column='_pl',
    table='',
)

danish_data_language = DataLocalizationLanguage(
    id=danish_language.id,
    name=danish_language.name,
    connection_id=xquery_datasource_connection.id,
)

# Add them languages to a project
project.data_language_settings.add_language(
    languages=[polish_data_language, danish_data_language]
)

# Convert Dutch language to SimpleLanguage object
dutch_metadata_language = SimpleLanguage(
    id=dutch_language.id,
    name=dutch_language.name,
)

# Add it to a project
project.metadata_language_settings.add_language(languages=[dutch_metadata_language])

# Remove languages from the project
project.data_language_settings.remove_language(
    languages=[polish_data_language, danish_data_language]
)
project.metadata_language_settings.remove_language(languages=[dutch_metadata_language])



import logging

from pandas import DataFrame, concat

from mstrio import config
from mstrio.api import documents, library
from mstrio.api.schedules import get_contents_schedule
from mstrio.connection import Connection
from mstrio.distribution_services.schedule import Schedule
from mstrio.helpers import IServerError
from mstrio.object_management import Folder, SearchPattern, search_operations
from mstrio.project_objects import OlapCube, SuperCube
from mstrio.server.environment import Environment
from mstrio.types import ObjectSubTypes
from mstrio.users_and_groups import User, UserGroup, UserOrGroup
from mstrio.utils import helper
from mstrio.utils.cache import CacheSource, ContentCacheMixin
from mstrio.utils.certified_info import CertifiedInfo
from mstrio.utils.entity import (
    CopyMixin,
    DeleteMixin,
    Entity,
    MoveMixin,
    ObjectTypes,
    VldbMixin,
)
from mstrio.utils.helper import (
    filter_params_for_func,
    find_object_with_name,
    get_valid_project_id,
    is_document,
)
from mstrio.utils.response_processors import objects as objects_processors
from mstrio.utils.translation_mixin import TranslationMixin
from mstrio.utils.version_helper import method_version_handler

logger = logging.getLogger(__name__)


def list_documents(
    connection: Connection,
    to_dictionary: bool = False,
    to_dataframe: bool = False,
    limit: int | None = None,
    name: str | None = None,
    project_id: str | None = None,
    project_name: str | None = None,
    **filters,
) -> list["Document"] | list[dict] | DataFrame:
    """Get all Documents available in the project specified within the
    `connection` object.

    Optionally use `to_dictionary` or `to_dataframe` to choose output format.
    If `to_dictionary` is True, `to_dataframe` is omitted.

    Args:
        connection (Connection): MicroStrategy connection object returned
            by 'connection.Connection()'
        to_dictionary (bool, optional): if True, return Documents as
            list of dicts
        to_dataframe (bool, optional): if True, return Documents as
            pandas DataFrame
        limit (int, optional): limit the number of elements returned.
            If `None` (default), all objects are returned.
        name (str, optional): characters that the document name must contain
        project_id (str, optional): Project ID
        project_name (str, optional): Project name
        **filters: Available filter parameters: ['name', 'id', 'type',
            'subtype', 'date_created', 'date_modified', 'version', 'acg',
            'owner', 'ext_type', 'view_media', 'certified_info', 'project_id']

    Returns:
            List of documents or list of dictionaries or DataFrame object
    """

    return Document._list_all(
        connection,
        to_dictionary=to_dictionary,
        name=name,
        limit=limit,
        to_dataframe=to_dataframe,
        project_id=project_id,
        project_name=project_name,
        **filters,
    )


def list_documents_across_projects(
    connection: Connection,
    name: str | None = None,
    to_dictionary: bool = False,
    to_dataframe: bool = False,
    limit: int | None = None,
    **filters,
) -> list["Document"] | list[dict] | DataFrame:
    """Get all Documents stored on the server.

    Optionally use `to_dictionary` or `to_dataframe` to choose output format.
    If `to_dictionary` is True, `to_dataframe` is omitted.

    Args:
        connection (Connection): MicroStrategy connection object returned
            by 'connection.Connection()'
        name (string, optional): characters that the document name must contain
        to_dictionary (bool, optional): if True, return Documents as
            list of dicts
        to_dataframe (bool, optional): if True, return Documents as
            pandas DataFrame
        limit (int, optional): limit the number of elements returned. If `None`
            (default), all objects are returned.
        **filters: Available filter parameters: ['name', 'id', 'type',
            'subtype', 'date_created', 'date_modified', 'version', 'acg',
            'owner', 'ext_type', 'view_media', 'certified_info', 'project_id']

    Returns:
            List of documents or list of dictionaries or DataFrame object
    """
    project_id_before = connection.project_id
    env = Environment(connection)
    projects = env.list_projects()
    output = DataFrame() if to_dataframe else []
    for project in projects:
        try:
            connection.select_project(project_id=project.id)
        except ValueError:
            if config.verbose:
                logger.info(
                    f'Project {project.name} ({project.id}) is skipped '
                    f'because it does not exist or user has no access '
                    f'to it'
                )
            continue

        docs = Document._list_all(
            connection,
            to_dictionary=to_dictionary,
            name=name,
            limit=limit,
            to_dataframe=to_dataframe,
            **filters,
        )
        if to_dataframe:
            output = concat([output, docs], ignore_index=True)
        else:
            output.extend(docs)

    connection.select_project(project_id=project_id_before)
    return output[:limit]


class Document(
    Entity,
    VldbMixin,
    CopyMixin,
    MoveMixin,
    DeleteMixin,
    ContentCacheMixin,
    TranslationMixin,
):
    """Python representation of MicroStrategy Document object

    _CACHE_TYPE is a variable used by ContentCache class for cache filtering
    purposes.
    """

    _OBJECT_TYPE = ObjectTypes.DOCUMENT_DEFINITION
    _CACHE_TYPE = CacheSource.Type.DOCUMENT
    _API_GETTERS = {**Entity._API_GETTERS, 'recipients': library.get_document}
    _API_PATCH = {
        ('name', 'description', 'folder_id', 'hidden'): (
            objects_processors.update,
            'partial_put',
        )
    }
    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        'owner': User.from_dict,
        'certified_info': CertifiedInfo.from_dict,
        'recipients': [User.from_dict],
    }

    def __init__(
        self, connection: Connection, name: str | None = None, id: str | None = None
    ):
        """Initialize Document object by passing name or id.

        Args:
            connection (object): MicroStrategy connection object returned
                by `connection.Connection()`
            name (string, optional): name of Document
            id (string, optional): ID of Document
        """
        if id is None:
            if name is None:
                raise ValueError(
                    "Please specify either 'name' or 'id' parameter in the constructor."
                )

            document = find_object_with_name(
                connection=connection,
                cls=self.__class__,
                name=name,
                listing_function=self._list_all,
                search_pattern=SearchPattern.EXACTLY,
            )
            id = document['id']
        super().__init__(connection=connection, object_id=id, name=name)

    def _init_variables(self, default_value, **kwargs) -> None:
        super()._init_variables(default_value=default_value, **kwargs)
        self._instance_id = ""
        self._recipients = kwargs.get('recipients', default_value)
        self._project_id = self.connection.project_id
        self._template_info = kwargs.get('templateInfo', default_value)
        self._folder_id = None

    def list_properties(self):
        """List properties for the document.

        Returns:
            A list of all document properties."""
        properties = {
            'id': self.id,
            'name': self.name,
            'description': self.description,
            'type': self.type,
            'subtype': self.subtype,
            'ext_type': self.ext_type,
            'ancestors': self.ancestors,
            'certified_info': self.certified_info,
            'comments': self.comments,
            'date_created': self.date_created,
            'date_modified': self.date_modified,
            'instance_id': self.instance_id,
            'owner': self.owner,
            'recipients': self.recipients,
            'version': self.version,
            'template_info': self.template_info,
            'acg': self.acg,
            'acl': self.acl,
        }
        return properties

    def alter(
        self,
        name: str | None = None,
        description: str | None = None,
        folder_id: Folder | str | None = None,
        hidden: bool | None = None,
    ):
        """Alter Document name, description and/or folder id.

        Args:
            name (string, optional): new name of the Document
            description (string, optional): new description of the Document
            folder_id (string | Folder, optional): A globally unique identifier
                used to distinguish between metadata objects within the same
                project. It is possible for two metadata objects in different
                projects to have the same Object ID.
            hidden: Specifies whether the document is hidden
        """
        description = description or self.description
        properties = filter_params_for_func(self.alter, locals(), exclude=['self'])
        self._alter_properties(**properties)
        if folder_id:
            self._folder_id = folder_id

    def __validate_user(self, recipient_id: str) -> str | None:
        try:
            User(self.connection, id=recipient_id)
        except IServerError:
            if config.verbose:
                logger.info(f'{recipient_id} is not a valid value for User ID')
            return None
        return recipient_id

    def publish(self, recipients: UserOrGroup | list[UserOrGroup] | None = None):
        """Publish the document for authenticated user. If `recipients`
        parameter is specified publishes the document for the given users.

        Args:
            recipients(UserOrGroup | list[UserOrGroup], optional): list of users
                or user groups to publish the document to (can be a list of IDs
                or a list of User and UserGroup elements)
        """
        if not isinstance(recipients, list) and recipients is not None:
            recipients = [recipients]

        if recipients is None:
            recipients = [self.connection.user_id]
        elif all(isinstance(el, User) for el in recipients):
            recipients = [recipient.id for recipient in recipients]
        elif all(isinstance(el, UserGroup) for el in recipients):
            users = [user for group in recipients for user in group.members]
            recipients = [user.id for user in users]
        elif any(not isinstance(el, str) for el in recipients):
            raise ValueError(
                'Please provide either list of User, UserGroup or str elements.'
            )
        for recipient in recipients:
            if not self.__validate_user(recipient):
                recipients.remove(recipient)  # noqa B038
        body = {'id': self.id, 'recipients': recipients}
        library.publish_document(self.connection, body)
        self.fetch(attr='recipients')

    def unpublish(self, recipients: UserOrGroup | list[UserOrGroup] | None = None):
        """Unpublish the document for all users it was previously published to.
        If `recipients` parameter is specified unpublishes the document for the
        given users.

        Args:
            recipients(UserOrGroup | list[UserOrGroup], optional): list of users
                or user groups to publish the document to (can be a list of IDs
                or a list of User and UserGroup elements)
        """

        if recipients is None:
            library.unpublish_document(self.connection, id=self.id)
        else:
            if not isinstance(recipients, list):
                recipients = [recipients]
            if all(isinstance(el, User) for el in recipients):
                recipients = [recipient.id for recipient in recipients]
            elif all(isinstance(el, UserGroup) for el in recipients):
                users = [user for group in recipients for user in group.members]
                recipients = [user.id for user in users]
            elif any(not isinstance(el, str) for el in recipients):
                raise ValueError(
                    'Please provide either list User and UserGroup elements or str '
                    'elements.'
                )
            for user_id in recipients:
                if self.__validate_user(user_id):
                    library.unpublish_document_for_user(
                        self.connection, document_id=self.id, user_id=user_id
                    )
        self.fetch(attr='recipients')

    @method_version_handler('11.3.0600')
    def list_available_schedules(
        self, to_dictionary: bool = False
    ) -> list["Schedule"] | list[dict]:
        """Get a list of schedules available for the object instance.

        Args:
            to_dictionary (bool, optional): If True returns a list of
                dictionaries, otherwise returns a list of Schedules.
                False by default.

        Returns:
            List of Schedule objects or list of dictionaries.
        """
        schedules_list_response = (
            get_contents_schedule(
                connection=self.connection,
                project_id=self.connection.project_id,
                body={'id': self.id, 'type': 'document'},
            ).json()
        ).get('schedules')
        if to_dictionary:
            return schedules_list_response
        else:
            return [
                Schedule.from_dict(connection=self.connection, source=schedule_id)
                for schedule_id in schedules_list_response
            ]

    def share_to(self, users: UserOrGroup | list[UserOrGroup]):
        """Shares the document to the listed users' libraries.

        Args:
            users(UserOrGroup | list[UserOrGroup]): list of users or user
                groups to publish the document to (can be a list of IDs or a
                list of User and UserGroup elements).
        """
        self.publish(users)

    @classmethod
    def _list_all(
        cls,
        connection: Connection,
        search_pattern: SearchPattern | int = SearchPattern.CONTAINS,
        name: str | None = None,
        to_dictionary: bool = False,
        to_dataframe: bool = False,
        limit: int | None = None,
        project_id: str | None = None,
        project_name: str | None = None,
        **filters,
    ) -> list["Document"] | list[dict] | DataFrame:
        if to_dictionary and to_dataframe:
            helper.exception_handler(
                "Please select either `to_dictionary=True` or `to_dataframe=True`, but "
                "not both.",
                ValueError,
            )
        project_id = get_valid_project_id(
            connection=connection,
            project_id=project_id,
            project_name=project_name,
            with_fallback=not project_name,
        )

        objects = search_operations.full_search(
            connection,
            object_types=ObjectSubTypes.REPORT_WRITING_DOCUMENT,
            project=project_id,
            name=name,
            pattern=search_pattern,
            **filters,
        )
        documents = [obj for obj in objects if is_document(obj['view_media'])]

        documents = documents[:limit] if limit else documents

        if to_dictionary:
            return documents
        elif to_dataframe:
            return DataFrame(documents)
        else:
            return [
                cls.from_dict(
                    source=document, connection=connection, with_missing_value=True
                )
                for document in documents
            ]

    def get_connected_cubes(self) -> list[SuperCube | OlapCube]:
        """Lists cubes used by this document.

        Returns:
            A list of cubes used by the document."""
        cubes = documents.get_cubes_used_by_document(self.connection, self.id).json()
        ret_cubes = [helper.choose_cube(self.connection, cube) for cube in cubes]
        return [tmp for tmp in ret_cubes if tmp]  # remove `None` values

    @property
    def instance_id(self):
        if self._instance_id == '':
            body = {"resolveOnly": True, "persistViewState": True}
            response = documents.create_new_document_instance(
                connection=self.connection, document_id=self.id, body=body
            )
            self._instance_id = response.json()['mid']
        return self._instance_id

    @property
    def folder_id(self):
        if not self._folder_id:
            self._folder_id = (
                next(folder['id'] for folder in self.ancestors if folder['level'] == 1)
                if self.ancestors
                else None
            )
        return self._folder_id

    @property
    def recipients(self):
        return self._recipients

    @property
    def template_info(self):
        return self._template_info



import logging
from enum import auto
from typing import Optional, Union

from mstrio.api import schema
from mstrio.connection import Connection
from mstrio.utils.entity import auto_match_args_entity
from mstrio.utils.enum_helper import AutoName, get_enum, get_enum_val
from mstrio.utils.helper import Dictable, exception_handler, get_valid_project_id
from mstrio.utils.time_helper import DatetimeFormats, map_str_to_datetime
from mstrio.utils.version_helper import class_version_handler

logger = logging.getLogger(__name__)


class SchemaLockType(AutoName):
    """Enumeration constants used to specify a type of lock that can be placed
    on schema. Type `UNLOCKED` is used only when displaying status.
    """

    ABSOLUTE_INDIVIDUAL = auto()
    EXCLUSIVE_CONSTITUENT = auto()
    ABSOLUTE_CONSTITUENT = auto()
    UNLOCKED = "none"


class SchemaUpdateType(AutoName):
    """Enumeration constants used to specify type of update for the schema."""

    TABLE_KEY = auto()
    ENTRY_LEVEL = auto()
    LOGICAL_SIZE = auto()
    CLEAR_ELEMENT_CACHE = auto()


class SchemaLockStatus(Dictable):
    """An object that contains all of the information about the lock status of
    the schema. If the schema is not locked then properties of the lock are not
    provided."""

    _FROM_DICT_MAP = {
        'lock_type': SchemaLockType,
        'date_created': DatetimeFormats.YMDHMS,
    }

    def __init__(
        self,
        lock_type: Union['SchemaLockType', str],
        date_created: str | None = None,
        comment: str | None = None,
        machine_name: str | None = None,
        owner_name: str | None = None,
        owner_id: str | None = None,
    ):
        """Initialize schema lock status. When schema is unlocked, only its
        `lock_type` is provided.

        Args:
            lock_type (str, enum): a type of lock status that could be returned
                from the schema. The lock state any of the lock types, but it is
                also possible that the schema is not actually locked. Available
                lock types are:
                - `absolute_individual`: When placed no one, including
                  administrator, can perform changes on the schema objects. The
                  purpose of this lock is to prevent accidental modifications to
                  the schema in a project.
                - `exclusive_constituent`: It is a lock exclusive to the
                  changeset holding the lock; only that changeset can change the
                  schema or its constituents (tables, attributes, etc.).
                - `absolute_constituent`: According to REST API documentation
                  this type is currently not in use.
                - `none`: Schema is unlocked.
                `lock_type` can be provided as string or enum `SchemaLockType`.
            date_created (str, optional): The date/time at which this lock was
                placed on the schema.
            comment (str, optional): Optional comment provided by the user that
                applied the lock. If used, this comment may help to justify why
                the schema was locked.
            machine_name (str, optional): Name of the machine used by the user
                that applied the lock.
            owner_name (str, optional): Name of the owner of the lock.
            owner_id (str, optional): ID of the owner of the lock.

        Raises:
            `TypeError` if `lock_type` is neither a string nor
            a `SchemaLockType`.
            `ValueError` if `lock_type` is specified as a string which is not
            a proper value of enum `SchemaLockType`.
        """

        self.lock_type = get_enum(lock_type, SchemaLockType)
        self.date_created = map_str_to_datetime(
            "date_created", date_created, self._FROM_DICT_MAP
        )
        self.comment = comment
        self.machine_name = machine_name
        self.owner_name = owner_name
        self.owner_id = owner_id


class SchemaTaskStatus(AutoName):
    """Enumeration constants used to specify status of the task."""

    RUNNING = auto()
    COMPLETED = auto()
    FAILED = auto()


class SchemaTaskError(Dictable):
    """Representation of properties used to report an error related to schema
    task."""

    def __init__(
        self, code: str, message: str, additional_properties: dict | None = None
    ):
        """Initialize task error.

        Args:
            code (str): Internal application error code.
            message (str): Description of error.
            additional_properties (dict, optional): Additional information
                related to the error (if any).
        """
        self.code = code
        self.message = message
        self.additional_properties = additional_properties


class SchemaTask(Dictable):
    """Detailed information about a task which is performed on the schema."""

    _FROM_DICT_MAP = {
        'status': SchemaTaskStatus,
        'start_time': DatetimeFormats.FULLDATETIME,
        'end_time': DatetimeFormats.FULLDATETIME,
        'errors': [SchemaTaskError],
    }

    def __init__(
        self,
        id: str,
        status: str | SchemaTaskStatus,
        start_time: str | None = None,
        end_time: str | None = None,
        errors: list[SchemaTaskError] | None = None,
    ):
        """Initialize task.

        Args:
            id (str): Task ID.
            status (str, enum): Task status. Possible values are defined in
                enum `SchemaTaskStatus`.
            start_time: (str, optional): The date/time at which the execution
                of task started.
            end_time (str, optional): The date/time at which the task execution
                completed (or failed). This value is not defined when task is
                still running.
            errors: (list, optional): List of errors if the task execution
                failed. Errors are represented by class `SchemaTaskError`.

        Raises:
            `TypeError` if `status` is neither a string nor
            a `SchemaTaskStatus`.
            `ValueError` if `status` is specified as a string which is not
            a proper value of enum `SchemaTaskStatus`.
        """
        self.id = id
        self.status = get_enum(status, SchemaTaskStatus)
        self.start_time = map_str_to_datetime(
            "start_time", start_time, self._FROM_DICT_MAP
        )
        self.end_time = map_str_to_datetime("end_time", end_time, self._FROM_DICT_MAP)
        self.errors = errors

    def __repr__(self) -> str:
        param_dict = auto_match_args_entity(
            self.__init__,
            self,
            exclude=['self', 'start_time', 'end_time', 'errors'],
            include_defaults=False,
        )

        params = [f'{param}={repr(value)}' for param, value in param_dict.items()]
        formatted_params = ", ".join(params)

        return f"SchemaTask({formatted_params})"


@class_version_handler('11.3.0000')
class SchemaManagement:
    """Representation of schema management object.

    Attributes:
        connection: instance of `Connection` object
        lock_type: type of lock which is placed on the schema
        project_id: ID of project on which schema is managed
        project_name: Project name
        tasks: array with objects of type `SchemaTask`. It represents
            tasks which were created for the current object of
            `SchemaManagement`. They are created when schema is reloaded
            asynchronously.
    """

    def __init__(
        self,
        connection: "Connection",
        project_id: str | None = None,
        project_name: str | None = None,
    ):
        """Initialize schema management object for the given project.

        Specify either `project_id` or `project_name`.
        When `project_id` is provided (not `None`), `project_name` is omitted.

        Note:
            When `project_id` is `None` and `project_name` is `None`, then
            its value is overwritten by `project_id` from `connection` object.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
            project_id (optional, str): ID of project on which schema will be
                managed.
            project_name (optional, str): Project name

        Raises:
            `AttributeError` when `project_id` is not provided and project is
            not selected in the `connection` object.
        """
        self.connection = connection
        self._project_id = get_valid_project_id(
            connection=connection,
            project_id=project_id,
            project_name=project_name,
            with_fallback=not project_name,
        )
        self._tasks = None
        self._lock_type = None

    def __repr__(self) -> str:
        param_dict = auto_match_args_entity(
            self.__init__, self, exclude=['self'], include_defaults=False
        )
        params = [
            (
                f"{param}"
                if (param == "connection" and isinstance(value, Connection))
                else f'{param}={repr(value)}'
            )
            for param, value in param_dict.items()
        ]
        formatted_params = ', '.join(params)

        return f'SchemaManagement({formatted_params})'

    def get_lock_status(self) -> "SchemaLockStatus":
        """Get the lock status of the schema.

        Returns:
            Lock status of the schema as `SchemaLockStatus` object.
        """
        res = schema.read_lock_status(self.connection, self.project_id).json()
        lock_status = SchemaLockStatus.from_dict(res)
        self._lock_type = lock_status.lock_type
        return lock_status

    def lock(self, lock_type: Union[str, "SchemaLockType"]) -> bool:
        """Lock the schema. After successfully locking the schema its
        `lock_type` property is updated.

        Args:
            lock_type: (string or object): Type of lock which will be placed on
                the schema.

        Returns:
            `True` when schema was successfully locked or schema has already
            been locked. `False` when procedure of locking schema failed.

        Raises:
            `TypeError` if `lock_type` is neither a string nor
            a `SchemaLockType`.
            `ValueError` if `lock_type` is specified as a string which is not
            a proper value of enum `SchemaLockType`.
        """
        lock_type = get_enum_val(lock_type, SchemaLockType)

        if self.lock_type != SchemaLockType.UNLOCKED:
            logger.info('Schema is already locked.')
            return True

        res = schema.lock_schema(self.connection, lock_type, self.project_id)
        if res.ok:
            self.get_lock_status()
            return True
        return False

    def unlock(self) -> bool:
        """Unlock the schema. After successfully unlocking the schema its
        `lock_type` property is updated.

        Returns:
            `True` when schema was successfully unlocked or it has already been
            unlocked. `False` when procedure of unlocking schema failed.
        """

        if self.lock_type == SchemaLockType.UNLOCKED:
            logger.info('Schema is already unlocked.')
            return True

        res = schema.unlock_schema(
            self.connection, project_id=self.project_id, throw_error=False
        )
        if res.ok:
            self.get_lock_status()
            return True
        return False

    def reload(
        self,
        update_types: (
            list[Union[str, "SchemaUpdateType"]] | Union[str, "SchemaUpdateType"] | None
        ) = None,
        respond_async: bool = True,
    ) -> Optional["SchemaTask"]:
        """Reload (update) the schema. This operation can be performed
        asynchronously. In that case the task is created and it is saved in
        property `tasks` to help tracking its status.

        Args:
            update_types (optional, list or object or string): Field with update
                type(s). Values in this field can be of type `string` or
                `SchemaUpdateType` or be a list of those types. This field can
                contain empty, any of the following options, or all of them:
                - `table_key`: Use this option if you changed the key structure
                  of a table.
                - `entry_level`: Use this option if you changed the level at
                  which a fact is stored.
                - `logical_size`: Use this option to recalculate logical table
                  sizes and override any modifications you made to logical table
                  sizes.
                - `clear_element_cache`: Use this option to clear up the
                  attribute element cache saved on the Intelligence Server.
            respond_async (optional, bool): When `True` reload is performed
                asynchronously. Otherwise it is performed synchronously. Default
                value is `True`.

        Returns:
            When `respond_async` is set to `True` then `SchemaTask` object with
            all details about the task of reloading schema is returned.
            Otherwise `None` is returned.

        Raises:
            `TypeError` if any value in `update_types` is neither a string nor
            a `SchemaUpdateType`.
            `ValueError` if any value in `update_types` is specified as a string
            which is not a proper value of enum `SchemaUpdateType`.
        """
        if not update_types:
            update_types = []
        elif not isinstance(update_types, list):
            update_types = [update_types]
        update_types = [get_enum_val(t, SchemaUpdateType) for t in update_types]

        res = schema.reload_schema(
            self.connection, self.project_id, update_types, respond_async
        )
        if res.ok and respond_async:
            return self._save_task(res.json())

    def get_task(self, task_index: int) -> "SchemaTask":
        """Get all details of the task which is stored at a given
        `task_index` in a list from property `tasks`.

        Args:
            task_index (int): Index of the task in the list stored in property
                `tasks`.

        Returns:
            `SchemaTask` object with all details about the task from the given
            index. When index is not proper then `None` is returned and warning
            with explanation message is shown.
        """
        try:
            task_id = self.tasks[task_index].id
        except IndexError:
            msg = (
                f"Cannot get task with index {task_index} from the list of tasks for "
                f"this schema management object. Check the list using property `tasks`."
            )
            exception_handler(msg, Warning)
            return

        res = schema.read_task_status(self.connection, task_id, self.project_id)
        return self._save_task(res.json())

    def _save_task(self, task_info: dict) -> 'SchemaTask':
        """Helper method to save the task in property `tasks` based on the
        argument `task_info` provided as a dictionary. When task with the given
        id was in the list before, then it is updated and its index in the list
        is not changed. Otherwise task is appended to the list."""
        task = SchemaTask.from_dict(task_info)
        if self._tasks is None:
            self._tasks = [task]
        else:
            # update task if it is already in the list ; otherwise append it
            try:
                index = [t.id for t in self._tasks].index(task.id)
                self._tasks[index] = task
            except ValueError:
                self._tasks.append(task)
        return task

    @property
    def tasks(self):
        """List with tasks created while running `reload` on the given instance
        of `SchemaManagement`."""
        if self._tasks is None:
            self._tasks = []
        return self._tasks

    @property
    def lock_type(self):
        """Current lock type of the schema."""
        if self._lock_type is None:
            self.get_lock_status()
        return self._lock_type

    @property
    def project_id(self):
        """Id of project which is managed by the `SchemaManagement` object."""
        return self._project_id
